@@ -269,7 +269,7 @@ void CallEndpoint::setLastPeerId()
 	return;
     Lock lock(s_mutex,5000000);
     if (!checkRetry(lock)) {
-	Alarm("engine","bug",DebugGoOn,"Set last peer ID failed - timeout on call endpoint mutex owned by '%s'!",s_mutex.owner());
+	Alarm("engine","bug",DebugCrit,"Set last peer ID failed - timeout on call endpoint mutex owned by '%s'!",s_mutex.owner());
 	return;
     }
     if (m_peer) {
@@ -290,7 +290,7 @@ DataEndpoint* CallEndpoint::getEndpoint(const String& type) const
 
 DataEndpoint* CallEndpoint::setEndpoint(const String& type)
 {
-    if (type.null())
+    if (type.null() || (refcount() <= 0))
 	return 0;
     DataEndpoint* dat = getEndpoint(type);
     if (!dat) {
@@ -303,7 +303,7 @@ DataEndpoint* CallEndpoint::setEndpoint(const String& type)
 
 void CallEndpoint::setEndpoint(DataEndpoint* endPoint)
 {
-    if (!(endPoint && endPoint->ref()))
+    if ((refcount() <= 0) || !(endPoint && endPoint->ref()))
 	return;
     if (m_data.find(endPoint)) {
 	endPoint->deref();
@@ -342,27 +342,27 @@ void CallEndpoint::clearEndpoint(const String& type)
 void CallEndpoint::setSource(DataSource* source, const String& type)
 {
     DataEndpoint* dat = source ? setEndpoint(type) : getEndpoint(type);
-    if (dat)
+    if (RefObject::alive(dat))
 	dat->setSource(source);
 }
 
 DataSource* CallEndpoint::getSource(const String& type) const
 {
     DataEndpoint* dat = getEndpoint(type);
-    return dat ? dat->getSource() : 0;
+    return RefObject::alive(dat) ? dat->getSource() : 0;
 }
 
 void CallEndpoint::setConsumer(DataConsumer* consumer, const String& type)
 {
     DataEndpoint* dat = consumer ? setEndpoint(type) : getEndpoint(type);
-    if (dat)
+    if (RefObject::alive(dat))
 	dat->setConsumer(consumer);
 }
 
 DataConsumer* CallEndpoint::getConsumer(const String& type) const
 {
     DataEndpoint* dat = getEndpoint(type);
-    return dat ? dat->getConsumer() : 0;
+    return RefObject::alive(dat) ? dat->getConsumer() : 0;
 }
 
 bool CallEndpoint::clearData(DataNode* node, const String& type)
@@ -475,7 +475,7 @@ void Channel::initChan()
     Lock mylock(m_driver);
 #ifndef NDEBUG
     if (m_driver->channels().find(this)) {
-	Debug(DebugGoOn,"Channel '%s' already in list of '%s' driver [%p]",
+	Debug(DebugCrit,"Channel '%s' already in list of '%s' driver [%p]",
 	    id().c_str(),m_driver->name().c_str(),this);
 	return;
     }
@@ -1011,6 +1011,14 @@ bool Channel::setDebug(Message& msg)
     if (str.startSkip("level")) {
 	int dbg = debugLevel();
 	str >> dbg;
+	if (str == "+") {
+	    if (debugLevel() > dbg)
+		dbg = debugLevel();
+	}
+	else if (str == "-") {
+	    if (debugLevel() < dbg)
+		dbg = debugLevel();
+	}
 	debugLevel(dbg);
     }
     else if (str == "reset")
@@ -1040,6 +1048,7 @@ TokenDict Module::s_messages[] = {
     { "engine.command",  Module::Command },
     { "engine.help",     Module::Help },
     { "engine.halt",     Module::Halt },
+    { "engine.stop",     Module::Stop },
     { "call.route",      Module::Route },
     { "call.execute",    Module::Execute },
     { "call.drop",       Module::Drop },
@@ -1307,6 +1316,14 @@ bool Module::setDebug(Message& msg, const String& target)
     if (str.startSkip("level")) {
 	int dbg = debugLevel();
 	str >> dbg;
+	if (str == "+") {
+	    if (debugLevel() > dbg)
+		dbg = debugLevel();
+	}
+	else if (str == "-") {
+	    if (debugLevel() < dbg)
+		dbg = debugLevel();
+	}
 	debugLevel(dbg);
     }
     else if (str == "reset") {
@@ -1349,7 +1366,8 @@ Driver::Driver(const char* name, const char* type)
       m_init(false), m_varchan(true),
       m_routing(0), m_routed(0), m_total(0),
       m_nextid(0), m_timeout(0),
-      m_maxroute(0), m_maxchans(0), m_chanCount(0), m_dtmfDups(false)
+      m_maxroute(0), m_maxchans(0), m_chanCount(0),
+      m_dtmfDups(false), m_doExpire(true)
 {
     m_prefix << name << "/";
 }
@@ -1410,20 +1428,25 @@ bool Driver::received(Message &msg, int id)
     String dest;
     switch (id) {
 	case Timer:
-	    {
-		// check each channel for timeouts
-		lock();
-		ListIterator iter(m_chans);
-		Time t;
-		for (;;) {
-		    RefPointer<Channel> c = static_cast<Channel*>(iter.get());
-		    unlock();
-		    if (!c)
-			break;
-		    c->checkTimers(msg,t);
-		    c = 0;
-		    lock();
+	    if (m_doExpire && lock(950000)) {
+		if (m_doExpire) {
+		    m_doExpire = false;
+		    // check each channel for timeouts
+		    ListIterator iter(m_chans);
+		    Time t;
+		    for (;;) {
+			RefPointer<Channel> c = static_cast<Channel*>(iter.get());
+			unlock();
+			if (!c)
+			    break;
+			c->checkTimers(msg,t);
+			c = 0;
+			lock();
+		    }
+		    m_doExpire = true;
 		}
+		else
+		    unlock();
 	    }
 	    return Module::received(msg,id);
 	case Status:
@@ -1611,7 +1634,9 @@ void Driver::statusDetail(String& str)
     ObjList* l = m_chans.skipNull();
     for (; l; l=l->skipNext()) {
 	Channel* c = static_cast<Channel*>(l->get());
-	str.append(c->id(),",") << "=" << c->status() << "|" << c->address() << "|" << c->getPeerId();
+	str.append(c->id(),",") << "=" << c->status()
+	    << "|" << String::uriEscape(c->address(),",;|"," +?&")
+	    << "|" << c->getPeerId();
     }
 }
 
@@ -1809,4 +1834,58 @@ void Router::cleanup()
     destruct(m_msg);
 }
 
+
+void CallAccount::pickAccountParams(const NamedList& params)
+{
+    NamedIterator iter(params);
+    Lock mylock(m_mutex);
+    m_inbParams.clearParams();
+    m_outParams.clearParams();
+    m_regParams.clearParams();
+    while (const NamedString* n = iter.get()) {
+	if (n->name().length() <= 4)
+	    continue;
+	String name = n->name().substr(4).trimSpaces();
+	if (n->name().startsWith("reg:"))
+	    m_regParams.setParam(name,*n);
+	else if (n->name().startsWith("inb:"))
+	    m_inbParams.setParam(name,*n);
+	else if (n->name().startsWith("out:"))
+	    m_outParams.setParam(name,*n);
+    }
+}
+
+void CallAccount::setInboundParams(NamedList& params)
+{
+    Lock mylock(m_mutex);
+    NamedIterator iter(m_inbParams);
+    while (const NamedString* n = iter.get()) {
+	String tmp(*n);
+	params.replaceParams(tmp);
+	params.setParam(n->name(),tmp);
+    }
+}
+
+void CallAccount::setOutboundParams(NamedList& params)
+{
+    Lock mylock(m_mutex);
+    NamedIterator iter(m_outParams);
+    while (const NamedString* n = iter.get()) {
+	String tmp(*n);
+	params.replaceParams(tmp);
+	params.setParam(n->name(),tmp);
+    }
+}
+
+void CallAccount::setRegisterParams(NamedList& params)
+{
+    Lock mylock(m_mutex);
+    NamedIterator iter(m_regParams);
+    while (const NamedString* n = iter.get()) {
+	String tmp(*n);
+	params.replaceParams(tmp);
+	params.setParam(n->name(),tmp);
+    }
+}
+
 /* vi: set ts=8 sw=4 sts=4 noet: */
