@@ -24,6 +24,7 @@
 
 #include <process.h>
 #include <shlobj.h>
+typedef void (*sig_func_t)(int);
 #define RTLD_NOW 0
 #define dlopen(name,flags) LoadLibrary(name)
 #define dlclose !FreeLibrary
@@ -49,6 +50,7 @@ __declspec(dllimport) BOOL WINAPI SHGetSpecialFolderPathA(HWND,LPSTR,INT,BOOL);
 #include <sys/wait.h>
 #include <sys/resource.h>
 typedef void* HMODULE;
+#define sig_func_t sighandler_t
 #define PATH_SEP "/"
 #ifndef CFG_DIR
 #ifdef HAVE_MACOSX_SUPPORT
@@ -139,7 +141,9 @@ using namespace TelEngine;
 
 // Supervisor control constants
 
-// Maximum the child's sanity pool can grow
+// Minimum configurable size of child's sanity pool
+#define MIN_SANITY 2
+// Default maximum the child's sanity pool can grow
 #define MAX_SANITY 5
 // Initial sanity buffer, allow some init time for the child
 #define INIT_SANITY 30
@@ -149,9 +153,19 @@ using namespace TelEngine;
 #define RUNDELAY_MAX 60000000
 // Amount we decerease delay towards minimum each time child proves sanity
 #define RUNDELAY_DEC 20000
+// Maximum supervised initial delay
+#define INITDELAY_MAX 60000
 // Size of log relay buffer in bytes
 #define MAX_LOGBUFF 4096
 
+// Minimum and maximum settable time jump
+#ifndef MIN_TIME_JUMP
+#define MIN_TIME_JUMP (MAX_SANITY + 3)
+#endif
+#ifndef MAX_TIME_JUMP
+#define MAX_TIME_JUMP 120
+#endif
+
 #ifndef HOST_NAME_MAX
 #define HOST_NAME_MAX 255
 #endif
@@ -165,12 +179,15 @@ static u_int64_t s_restarts = 0;
 static bool s_makeworker = true;
 static bool s_keepclosing = false;
 static bool s_nounload = false;
+static long s_timejump = 0;
 static int s_super_handle = -1;
 static int s_run_attempt = 0;
+static int s_max_sanity = MAX_SANITY;
 static bool s_interactive = true;
 static bool s_localsymbol = false;
 static bool s_logtruncate = false;
 static const char* s_logfile = 0;
+static sig_func_t s_abrt_handler = SIG_DFL;
 
 static void sighandler(int signal)
 {
@@ -207,6 +224,36 @@ static void sighandler(int signal)
     }
 }
 
+static void abrthandler(int sig)
+{
+    ::signal(SIGABRT,s_abrt_handler);
+    Engine* eng = Engine::self();
+    if (eng) {
+	// TODO: retrieve all the information without getting any mutex locks
+	uint64_t enq,deq,disp,qmax;
+	eng->getStats(enq,deq,disp,qmax);
+	::fprintf(stderr,
+	    "*** ABORT INFO BEGIN ***\n"
+	    "  Messages:    " FMT64U "\n"
+	    "  MaxQueue:    " FMT64U "\n"
+	    "  Msg Age:     %u\n"
+	    "  Msg Rate:    %u\n"
+	    "  Max Rate:    %u\n"
+	    "  Threads:     %d\n"
+	    "  Workers:     %d\n"
+	    "  Mutexes:     %d\n"
+	    "  Locks:       %d\n"
+	    "  Congestion:  %u\n"
+	    "  Call Accept: %s\n"
+	    "*** ABORT INFO END ***\n",
+	    (enq - deq),qmax,eng->messageAge(),eng->messageRate(),eng->messageMaxRate(),
+	    Thread::count(),EnginePrivate::count,Mutex::count(),Mutex::locks(),
+	    Engine::getCongestion(),lookup(Engine::accept(),Engine::getCallAcceptStates())
+	);
+    }
+    ::raise(SIGABRT);
+}
+
 String Engine::s_node;
 String Engine::s_shrpath(SHR_PATH);
 String Engine::s_cfgsuffix(CFG_SUFFIX);
@@ -218,6 +265,7 @@ NamedList Engine::s_params("");
 Engine::RunMode Engine::s_mode = Engine::Stopped;
 Engine::CallAccept Engine::s_accept = Engine::Accept;
 Engine* Engine::s_self = 0;
+bool Engine::s_started = false;
 int Engine::s_haltcode = -1;
 int EnginePrivate::count = 0;
 static String s_cfgpath(CFG_PATH);
@@ -226,7 +274,13 @@ static bool s_createusr = true;
 static bool s_init = false;
 static bool s_dynplugin = false;
 static Engine::PluginMode s_loadMode = Engine::LoadFail;
+static int s_minworkers = 1;
 static int s_maxworkers = 10;
+static int s_addworkers = 1;
+static int s_maxmsgrate = 0;
+static int s_maxmsgage = 0;
+static int s_maxqueued = 0;
+static int s_exit = -1;
 unsigned int Engine::s_congestion = 0;
 static Mutex s_congMutex(false,"Congestion");
 static bool s_debug = true;
@@ -477,7 +531,13 @@ bool EngineStatusHandler::received(Message &msg)
     msg.retValue() << ",inuse=" << Engine::self()->usedPlugins();
     msg.retValue() << ",handlers=" << Engine::self()->handlerCount();
     msg.retValue() << ",hooks=" << Engine::self()->postHookCount();
-    msg.retValue() << ",messages=" << Engine::self()->messageCount();
+    uint64_t enq,deq,disp,qmax;
+    Engine::self()->getStats(enq,deq,disp,qmax);
+    msg.retValue() << ",messages=" << (enq - deq) << ",maxqueue=" << qmax;
+    msg.retValue() << ",messageage=" << Engine::self()->messageAge();
+    msg.retValue() << ",messagerate=" << Engine::self()->messageRate();
+    msg.retValue() << ",maxmsgrate=" << Engine::self()->messageMaxRate();
+    msg.retValue() << ",enqueued=" << enq << ",dequeued=" << deq << ",dispatched=" << disp ;
     msg.retValue() << ",supervised=" << (s_super_handle >= 0);
     msg.retValue() << ",runattempt=" << s_run_attempt;
 #ifndef _WINDOWS
@@ -1033,10 +1093,10 @@ static void copystream(int dest, int src)
     rotatelogs();
 }
 
-static int supervise(void)
+static int supervise(int initDelay)
 {
     s_superpid = ::getpid();
-    ::fprintf(stderr,"Supervisor (%u) is starting\n",s_superpid);
+    ::fprintf(stderr,"Supervisor (%d) is starting, max sanity %d\n",s_superpid,s_max_sanity);
     ::signal(SIGINT,superhandler);
     ::signal(SIGTERM,superhandler);
     ::signal(SIGHUP,superhandler);
@@ -1045,6 +1105,11 @@ static int supervise(void)
     ::signal(SIGUSR1,superhandler);
     ::signal(SIGUSR2,superhandler);
     ::signal(SIGALRM,SIG_IGN);
+    if (initDelay > 0) {
+	::fprintf(stderr,"Supervisor (%d) delaying initial start by %u.%02u seconds\n",
+		s_superpid,(initDelay + 5) / 1000,((initDelay + 5) / 10) % 100);
+	::usleep(initDelay * 1000);
+    }
     int retcode = 0;
     while (s_runagain) {
 	int wdogfd[2];
@@ -1126,8 +1191,8 @@ static int supervise(void)
 	    if (tmp >= 0) {
 		// Timer messages add one sanity point every second
 		tmp += sanity;
-		if (tmp > MAX_SANITY)
-		    tmp = MAX_SANITY;
+		if (tmp > s_max_sanity)
+		    tmp = s_max_sanity;
 		if (sanity < tmp)
 		    sanity = tmp;
 		// Decrement inter-run delay each time child proves sanity
@@ -1223,7 +1288,7 @@ SLib::~SLib()
 	if (fini || m_nounload) {
 	    count -= plugins.count();
 	    if (count != m_count)
-		Debug(DebugGoOn,"Finalizing '%s' removed %u out of %u plugins",
+		Debug(DebugCrit,"Finalizing '%s' removed %u out of %u plugins",
 		    c_str(),count,m_count);
 	    checkPoint();
 	    return;
@@ -1232,19 +1297,19 @@ SLib::~SLib()
     }
     int err = dlclose(m_handle);
     if (err)
-	Debug(DebugGoOn,"Error %d on dlclose(%p) of '%s'",err,m_handle,c_str());
+	Debug(DebugCrit,"Error %d on dlclose(%p) of '%s'",err,m_handle,c_str());
     else if (s_keepclosing) {
 	int tries;
 	for (tries=0; tries<10; tries++)
 	    if (dlclose(m_handle))
 		break;
 	if (tries)
-	    Debug(DebugGoOn,"Made %d attempts to dlclose(%p) '%s'",
+	    Debug(DebugCrit,"Made %d attempts to dlclose(%p) '%s'",
 		tries,m_handle,c_str());
     }
     count -= plugins.count();
     if (count != m_count)
-	Debug(DebugGoOn,"Unloading '%s' removed %u out of %u plugins",
+	Debug(DebugWarn,"Unloading '%s' removed %u out of %u plugins",
 	    c_str(),count,m_count);
     checkPoint();
 }
@@ -1342,6 +1407,8 @@ unsigned int SharedVars::dec(const String& name, unsigned int wrap)
 
 
 Engine::Engine()
+    : m_dispatchedLast(0), m_messageRate(0), m_maxMsgRate(0),
+      m_rateCongested(false), m_queueCongested(false), m_ageCongested(false)
 {
     DDebug(DebugAll,"Engine::Engine() [%p]",this);
     initUsrPath(s_usrpath);
@@ -1369,7 +1436,7 @@ int Engine::engineInit()
     WSADATA wsaData;
     int errc = ::WSAStartup(MAKEWORD(2,2), &wsaData);
     if (errc) {
-	Debug(DebugGoOn,"Failed to initialize the Windows Sockets library, error code %d",errc);
+	Debug(DebugCrit,"Failed to initialize the Windows Sockets library, error code %d",errc);
 	return errc & 127;
     }
 #else
@@ -1427,9 +1494,18 @@ int Engine::engineInit()
     const char *modPath = s_cfg.getValue("general","modpath");
     if (modPath)
 	s_modpath = modPath;
-    s_maxworkers = s_cfg.getIntValue("general","maxworkers",s_maxworkers);
-    s_maxevents = s_cfg.getIntValue("general","maxevents",s_maxevents);
+    s_minworkers = s_cfg.getIntValue("general","minworkers",s_minworkers,1,100);
+    s_maxworkers = s_cfg.getIntValue("general","maxworkers",s_maxworkers,s_minworkers,500);
+    s_addworkers = s_cfg.getIntValue("general","addworkers",s_addworkers,1,10);
+    s_maxmsgrate = s_cfg.getIntValue("general","maxmsgrate",s_maxmsgrate,0,50000);
+    s_maxmsgage = s_cfg.getIntValue("general","maxmsgage",s_maxmsgage,0,5000);
+    s_maxqueued = s_cfg.getIntValue("general","maxqueued",s_maxqueued,0,10000);
+    s_maxevents = s_cfg.getIntValue("general","maxevents",s_maxevents,0,1000);
     s_restarts = s_cfg.getIntValue("general","restarts");
+    s_timejump = s_cfg.getIntValue("general","timejump",0,0,MAX_TIME_JUMP);
+    if (s_timejump && (s_timejump < MIN_TIME_JUMP))
+	s_timejump = MIN_TIME_JUMP;
+    s_timejump *= 1000;
     m_dispatcher.warnTime(1000*(u_int64_t)s_cfg.getIntValue("general","warntime"));
     extraPath(clientMode() ? "client" : "server");
     extraPath(s_cfg.getValue("general","extrapath"));
@@ -1454,7 +1530,12 @@ int Engine::engineInit()
 #ifndef _WINDOWS
     s_params.addParam("lastsignal",String(s_childsig));
 #endif
+    s_params.addParam("minworkers",String(s_minworkers));
     s_params.addParam("maxworkers",String(s_maxworkers));
+    s_params.addParam("addworkers",String(s_addworkers));
+    s_params.addParam("maxmsgrate",String(s_maxmsgrate));
+    s_params.addParam("maxmsgage",String(s_maxmsgage));
+    s_params.addParam("maxqueued",String(s_maxqueued));
     s_params.addParam("maxevents",String(s_maxevents));
     if (track)
 	s_params.addParam("trackparam",track);
@@ -1497,12 +1578,15 @@ int Engine::engineInit()
 	Debug(DebugWarn,"No supervisor - disabling automatic restarts");
 	s_restarts = 0;
     }
+    if (s_cfg.getBoolValue("general","abortinfo",true))
+	s_abrt_handler = ::signal(SIGABRT,abrthandler);
     initPlugins();
     checkPoint();
     ::signal(SIGINT,sighandler);
     ::signal(SIGTERM,sighandler);
     Debug(DebugAll,"Engine dispatching start message");
     dispatch("engine.start",true);
+    s_started = true;
     internalStatisticsStart();
     setStatus(SERVICE_RUNNING);
 #ifndef _WINDOWS
@@ -1529,6 +1613,7 @@ int Engine::run()
 {
     // engine loop
     long corr = 0;
+    uint64_t last = 0;
     int stops = MAX_STOP;
     while (s_haltcode == -1 || ((--stops >= 0) && dispatch("engine.stop",true))) {
 	if (s_cmds) {
@@ -1549,6 +1634,23 @@ int Engine::run()
 
 	if (s_init) {
 	    s_init = false;
+	    s_cfg.load();
+	    s_params.setParam("maxworkers",String((s_maxworkers
+		= s_cfg.getIntValue("general","maxworkers",s_maxworkers,s_minworkers,1000))));
+	    s_params.setParam("addworkers",String((s_addworkers
+		= s_cfg.getIntValue("general","addworkers",s_addworkers,1,10))));
+	    s_params.setParam("maxmsgrate",String((s_maxmsgrate
+		= s_cfg.getIntValue("general","maxmsgrate",s_maxmsgrate,0,50000))));
+	    s_params.setParam("maxmsgage",String((s_maxmsgage
+		= s_cfg.getIntValue("general","maxmsgage",s_maxmsgage,0,5000))));
+	    s_params.setParam("maxqueued",String((s_maxqueued
+		= s_cfg.getIntValue("general","maxqueued",s_maxqueued,0,10000))));
+	    s_params.setParam("maxevents",String((s_maxevents
+		= s_cfg.getIntValue("general","maxevents",s_maxevents,0,1000))));
+	    s_timejump = s_cfg.getIntValue("general","timejump",s_timejump,0,MAX_TIME_JUMP);
+	    if (s_timejump && (s_timejump < MIN_TIME_JUMP))
+		s_timejump = MIN_TIME_JUMP;
+	    s_timejump *= 1000;
 	    initPlugins();
 	}
 
@@ -1575,27 +1677,74 @@ int Engine::run()
 	    CapturedEvent::capturing(false);
 	}
 
+	if (s_exit >= 0) {
+	    halt(s_exit);
+	    s_exit = -1;
+	}
+
 	// Create worker thread if we didn't hear about any of them in a while
-	if (s_makeworker && (EnginePrivate::count < s_maxworkers)) {
-	    if (EnginePrivate::count)
-		Alarm("engine","performance",(EnginePrivate::count < 4) ? DebugMild : DebugWarn,
-		    "Creating new message dispatching thread (%d running)",EnginePrivate::count);
-	    else
-		Debug(DebugInfo,"Creating first message dispatching thread");
-	    EnginePrivate *prv = new EnginePrivate;
-	    prv->startup();
+	int build = s_maxworkers - EnginePrivate::count;
+	if (s_makeworker && (build > 0)) {
+	    if (EnginePrivate::count) {
+		if (build > s_addworkers)
+		    build = s_addworkers;
+		Alarm("engine","performance",
+		    (EnginePrivate::count > (s_minworkers + 3)) ? DebugWarn : DebugMild,
+		    "Creating new %d message dispatching threads (%d running)",
+		    build,EnginePrivate::count);
+	    }
+	    else {
+		build = s_minworkers;
+		Debug(DebugInfo,"Creating first %d message dispatching threads",build);
+	    }
+	    do {
+		(new EnginePrivate)->startup();
+	    } while (--build > 0);
 	}
 	else
 	    s_makeworker = true;
 
-	if (s_restarts && (Time::now() >= s_restarts)) {
+	uint64_t now = Time::now();
+	if (last) {
+	    long diff = (long)((int64_t)(now - last) / 1000);
+	    if ((diff <= 0) || (s_timejump && (diff > s_timejump))) {
+		const char* rst = (s_timejump && (s_super_handle >= 0)) ? " Restarting!" : "";
+		Alarm("engine","system",DebugCrit,"System time advanced by %ld msec.%s",diff,rst);
+		if (s_timejump)
+		    restart(0,false);
+	    }
+	}
+	last = now;
+	if (s_restarts && (now >= s_restarts)) {
 	    if (!(usedPlugins() || dispatch("engine.busy"))) {
-		s_haltcode = 128;
+		s_restarts = 0;
+		restart(0,false);
 		break;
 	    }
 	    DDebug(DebugAll,"Engine busy - will try to restart later");
 	    // If we cannot restart now try again in 10s
-	    s_restarts = Time::now() + 10000000;
+	    s_restarts = now + 10000000;
+	}
+
+	uint64_t disp = m_dispatcher.dispatchCount();
+	m_messageRate = (unsigned int)(disp - m_dispatchedLast);
+	m_dispatchedLast = disp;
+	if (m_maxMsgRate < m_messageRate)
+	    m_maxMsgRate = m_messageRate;
+	bool cong = s_maxmsgrate && (m_messageRate > (unsigned)s_maxmsgrate);
+	if (cong != m_rateCongested) {
+	    m_rateCongested = cong;
+	    setCongestion(cong ? "message rate over limit" : 0);
+	}
+	cong = s_maxmsgage && (m_dispatcher.messageAge() > (unsigned)s_maxmsgage);
+	if (cong != m_ageCongested) {
+	    m_ageCongested = cong;
+	    setCongestion(cong ? "message age over limit" : 0);
+	}
+	cong = s_maxqueued && (m_dispatcher.messageCount() > (unsigned)s_maxqueued);
+	if (cong != m_queueCongested) {
+	    m_queueCongested = cong;
+	    setCongestion(cong ? "message queue over limit" : 0);
 	}
 
 	// Attempt to sleep until the next full second
@@ -1658,7 +1807,7 @@ int Engine::engineCleanup()
     unsigned int cnt = plugins.count();
     plugins.clear();
     if (mux || cnt)
-	Debug(DebugGoOn,"Exiting with %d locked mutexes and %u plugins loaded!",mux,cnt);
+	Debug(DebugWarn,"Exiting with %d locked mutexes and %u plugins loaded!",mux,cnt);
     if (GenObject::getObjCounting()) {
 	String str;
 	int obj = EngineStatusHandler::objects(str);
@@ -1702,6 +1851,9 @@ void Engine::setCongestion(const char* reason)
 	    if (reason)
 		Alarm("engine","performance",DebugWarn,"Engine is congested: %s",reason);
 	    break;
+	default:
+	    if (reason)
+		Debug("engine",DebugNote,"Engine extra congestion: %s",reason);
     }
 }
 
@@ -1811,7 +1963,10 @@ bool Engine::loadPluginDir(const String& relPath)
 #endif
     bool defload = s_cfg.getBoolValue("general","modload",true);
     String path = s_modpath;
-    if (relPath) {
+    static const Regexp r("^\\([/\\]\\|[[:alpha:]]:[/\\]\\).");
+    if (r.matches(relPath))
+	path = relPath;
+    else if (relPath) {
 	if (!path.endsWith(PATH_SEP))
 	    path += PATH_SEP;
 	path += relPath;
@@ -2229,7 +2384,7 @@ int Engine::cleanupLibrary()
     if (mux < 0)
 	mux = 0;
     if (mux)
-	Debug(DebugGoOn,"Exiting with %d locked mutexes!",mux);
+	Debug(DebugCrit,"Exiting with %d locked mutexes!",mux);
     if (GenObject::getObjCounting()) {
 	String str;
 	int obj = EngineStatusHandler::objects(str);
@@ -2255,7 +2410,7 @@ static void usage(bool client, FILE* f)
 "   -c pathname    Path to conf files directory (" CFG_PATH ")\n"
 "   -u pathname    Path to user files directory (%s)\n"
 "   -m pathname    Path to modules directory (" MOD_PATH ")\n"
-"   -x relpath     Relative path to extra modules directory (can be repeated)\n"
+"   -x dirpath     Absolute or relative path to extra modules directory (can be repeated)\n"
 "   -w directory   Change working directory\n"
 "   -N nodename    Set the name of this node in a cluster\n"
 #ifdef RLIMIT_CORE
@@ -2294,8 +2449,9 @@ static void usage(bool client, FILE* f)
 "   --remove       Remove the Windows service\n"
 #else
 "   -d             Daemonify, suppress output unless logged\n"
-"   -s             Supervised, restart if crashes or locks up\n"
+"   -s[=msec]      Supervised, restart if crashes or locks up, optionally sleeps initially\n"
 "   -r             Enable rotation of log file (needs -s and -l)\n"
+"   -S sanity      Set size of supervised sanity points pool\n"
 #endif
     ,s_cfgfile.safe()
     ,s_usrpath.safe());
@@ -2327,7 +2483,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, En
     int service = 0;
 #else
     bool daemonic = false;
-    bool supervised = false;
+    int supervised = 0;
 #endif
     bool client = (mode == Client);
     Debugger::Formatting tstamp = Debugger::TextLSep;
@@ -2420,11 +2576,29 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, En
 			daemonic = true;
 			break;
 		    case 's':
-			supervised = true;
+			supervised = -1;
+			if ('=' == pc[1]) {
+			    long int ms = ::strtol(pc+2,0,0);
+			    if (ms > INITDELAY_MAX)
+				supervised = INITDELAY_MAX;
+			    else if (ms > 0)
+				supervised = ms;
+			    pc = 0;
+			}
 			break;
 		    case 'r':
 			s_logrotator = true;
 			break;
+		    case 'S':
+			GET_PARAM;
+			s_max_sanity = ::strtol(param,0,0);
+			if (s_max_sanity <= 0)
+			    s_max_sanity = MAX_SANITY;
+			else if (s_max_sanity < MIN_SANITY)
+			    s_max_sanity = MIN_SANITY;
+			else if (s_max_sanity > INIT_SANITY)
+			    s_max_sanity = INIT_SANITY;
+			break;
 #endif
 		    case 't':
 			s_logtruncate = true;
@@ -2509,7 +2683,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, En
 				    s_init = true;
 				    break;
 				case 'x':
-				    s_haltcode++;
+				    s_exit++;
 				    break;
 				case 'w':
 				    s_makeworker = false;
@@ -2736,7 +2910,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, En
     int retcode = -1;
 #ifndef _WINDOWS
     if (supervised)
-	retcode = supervise();
+	retcode = supervise(supervised);
     if (retcode >= 0)
 	return retcode;
 #endif
