@@ -25,6 +25,15 @@
 #include <stdio.h>
 #include <regex.h>
 
+#if (defined(WORDS_BIGENDIAN) || defined(BIGENDIAN))
+#define ENDIANNESS_NATIVE (UChar::BE)
+#define ENDIANNESS_OPPOSITE (UChar::LE)
+#else
+#define ENDIANNESS_NATIVE (UChar::LE)
+#define ENDIANNESS_OPPOSITE (UChar::BE)
+#endif
+
+
 namespace TelEngine {
 
 // String to regular integer conversion, takes into account overflows
@@ -251,6 +260,145 @@ bool UChar::decode(const char*& str, uint32_t maxChar, bool overlong)
     return true;
 }
 
+static inline uint16_t swap_u16(uint16_t val, UChar::Endianness order)
+{
+    if (order < UChar::Native && ENDIANNESS_OPPOSITE == order)
+	val = ((val & 0xff00) >> 8) | ((val & 0x00ff) << 8);
+    return val;
+}
+
+bool UChar::decode(uint16_t*& buff, unsigned int& len, Endianness order, uint32_t maxChar)
+{
+    operator=('\0');
+    if (!(buff && len))
+	return false;
+    if (maxChar < 128)
+	maxChar = 0x10ffff; // RFC 3629 default limit
+    uint32_t val = swap_u16(*buff,order);
+    buff++;
+    len--;
+    if (val >= 0xD800 && val < 0xDC00 && len) { // High surrogate
+	uint16_t low = swap_u16(*buff,order);
+	if (low >= 0xDC00 && low <= 0xDFFF) {
+	    buff++;
+	    len--;
+	    val = (low - 0xDC00) + (val - 0xD800) * 0x400 + 0x10000;
+	}
+    }
+    operator=(val);
+    if (code() > maxChar)
+	return false;
+    return true;
+}
+
+bool UChar::decode(DataBlock& buff, Endianness order, uint32_t maxChar)
+{
+    operator=('\0');
+    unsigned int len = buff.length();
+    uint16_t* in = (uint16_t*) buff.data();
+    if (!len || (len & 1))
+	return false;
+    len = len >> 1;
+    if (!decode(in,len,order,maxChar))
+	return false;
+    buff.cut(-(buff.length() - len * 2));
+    return true;
+}
+
+bool UChar::encode(uint16_t*& buff, unsigned int& len, Endianness order)
+{
+    if (!(buff && len && code() <= 0x10ffff))
+	return false;
+    XDebug(DebugAll,"UChar::encode() UTF-16, char=%s (%x), order=%u",c_str(),m_chr,order);
+    if (m_chr >= 0x10000) { // encode to surrogate pairs
+	if (len < 2)
+	    return false; // not enough space to encode
+	*buff = swap_u16(((m_chr - 0x10000) >> 10) + 0xD800,order);
+	*(buff + 1) = swap_u16(((m_chr - 0x10000) & 0x3ff) + 0xDC00,order);
+	buff += 2;
+	len -=2;
+    }
+    else {
+	*buff = swap_u16(m_chr,order);
+	buff++;
+	len--;
+    }
+    return true;
+}
+
+bool UChar::encode(DataBlock& buff, Endianness order)
+{
+    uint16_t b[2] = {0};
+    uint16_t* out = b;
+    unsigned int len = 2;
+    if (!encode(out,len,order))
+	return false;
+    buff.append(b,sizeof(uint16_t) * (2 - len));
+    return true;
+}
+
+bool UChar::decode(String& out, uint16_t*& buff, unsigned int& len, Endianness order, bool checkBOM, uint32_t maxChar)
+{
+    if (!(buff && len))
+	return false;
+    XDebug(DebugAll,"UChar::decode() UTF-16, out=%s, buff=%p, len=%u, order=%u, maxChar=%x",
+	    out.c_str(),buff,len,order,maxChar);
+    if (checkBOM && (*buff == 0xfeff || *buff == 0xfffe)) {
+	if (*buff == 0xfeff)  // same endianness
+	    order = ENDIANNESS_NATIVE;
+	else
+	    order = ENDIANNESS_OPPOSITE;
+	buff++;
+	len--;
+    }
+    while (buff && len) {
+	UChar c;
+	if (!c.decode(buff,len,(Endianness)order,maxChar))
+	    return false;
+	out << c;
+    }
+    return true;
+}
+
+bool UChar::encode(DataBlock& out, const char*& str, Endianness order, bool addBOM)
+{
+    XDebug(DebugAll,"UChar::encode() UTF-16, str=%s, order=%u, addBOM=%s",str,order,String::boolText(addBOM));
+    if (TelEngine::null(str))
+	return false;
+    if (addBOM) {
+	uint16_t bom = swap_u16(0xfeff,order);
+	out.append(&bom,2);
+    }
+    UChar c;
+    while (*str && c.decode(str)) {
+	if (!c.encode(out,order))
+	    return false;
+    }
+    return true;
+}
+
+bool UChar::encode(uint16_t*& buff, unsigned int& len, const char*& str, Endianness order, bool addBOM)
+{
+    if (TelEngine::null(str))
+	return false;
+    if (!(buff && len))
+	return false;
+    XDebug(DebugAll,"UChar::encode() UTF-16, buff=%p, len=%u, str=%s, order=%u, addBOM=%s",
+	  buff,len,str,order,String::boolText(addBOM));
+    if (addBOM) {
+	uint16_t bom = swap_u16(0xfeff,order);
+	*buff = bom;
+	++buff;
+	len--;
+    }
+    UChar c;
+    while (*str && c.decode(str)) {
+	if (!c.encode(buff,len,order))
+	    return false;
+    }
+    return true;
+}
+
 
 StringMatchPrivate::StringMatchPrivate()
 {
@@ -1024,7 +1172,7 @@ String& String::append(double value, unsigned int decimals)
     return operator+=(buf);
 }
 
-static char* string_printf(unsigned int length, const char* format, va_list& va)
+static char* string_printf(unsigned int& length, const char* format, va_list& va)
 {
     if (TelEngine::null(format) || !length)
 	return 0;
@@ -1033,8 +1181,41 @@ static char* string_printf(unsigned int length, const char* format, va_list& va)
 	Debug("String",DebugFail,"malloc(%d) returned NULL!",length);
 	return 0;
     }
+    // Remember vsnprintf:
+    // standard:
+    //  - buffer size and returned value include the terminating NULL char
+    //  - returns -1 on error
+    // Windows:
+    //  - it doesn't write terminating NULL if there is no space
+    //  - returns -1 on error or buffer length is less than number of bytes to write
+    //  - returned value doesn't include the terminating NULL char (even if it was written to output)
+    buf[length] = 0;
+#ifdef _WINDOWS
+    errno = 0;
     int len = ::vsnprintf(buf,length,format,va);
-    buf[len] = 0;
+#else
+    int len = ::vsnprintf(buf,length + 1,format,va);
+#endif
+    if (len < 0) {
+#ifdef _WINDOWS
+	if (errno == ERANGE) {
+	    XDebug("String",DebugCrit,"string_printf() incomplete write");
+	    buf[length] = 0;
+	    length = 0;
+	    return buf;
+	}
+#endif
+	::free(buf);
+	Debug("String",DebugCrit,"string_printf(): vsnprintf() failed!");
+	return 0;
+    }
+    if (len < (int)length)
+	length = len;
+#ifdef XDEBUG
+    else if (len > (int)length || buf[len])
+	Debug("String",DebugCrit,"string_printf() incomplete write");
+#endif
+    buf[length] = 0;
     return buf;
 }
 
@@ -1050,6 +1231,7 @@ String& String::printf(unsigned int length, const char* format,  ...)
     }
     char* old = m_string;
     m_string = buf;
+    m_length = length;
     ::free(old);
     changed();
     return *this;
@@ -1068,6 +1250,7 @@ String& String::printf(const char* format, ...)
     }
     char* old = m_string;
     m_string = buf;
+    m_length = len;
     ::free(old);
     changed();
     return *this;
@@ -1416,8 +1599,24 @@ String String::uriEscape(const char* str, char extraEsc, const char* noEsc)
 	return s;
     char c;
     while ((c=*str++)) {
-	if ((unsigned char)c <= ' ' || c == '%' || c == extraEsc ||
-	    ((c == '+' || c == '?' || c == '&') && !(noEsc && ::strchr(noEsc,c))))
+	if ((unsigned char)c < ' ' || c == '%' || c == extraEsc ||
+	    ((c == ' ' || c == '+' || c == '?' || c == '&') && !(noEsc && ::strchr(noEsc,c))))
+	    s << '%' << hexEncode(c >> 4) << hexEncode(c);
+	else
+	    s += c;
+    }
+    return s;
+}
+
+String String::uriEscape(const char* str, const char* extraEsc, const char* noEsc)
+{
+    String s;
+    if (TelEngine::null(str))
+	return s;
+    char c;
+    while ((c=*str++)) {
+	if ((unsigned char)c < ' ' || c == '%' || (extraEsc && ::strchr(extraEsc,c)) ||
+	    ((c == ' ' || c == '+' || c == '?' || c == '&') && !(noEsc && ::strchr(noEsc,c))))
 	    s << '%' << hexEncode(c >> 4) << hexEncode(c);
 	else
 	    s += c;
@@ -1681,13 +1880,13 @@ const String* String::atom(const String*& str, const char* val)
 
 
 Regexp::Regexp()
-    : m_regexp(0), m_flags(0)
+    : m_regexp(0), m_compile(true), m_flags(0)
 {
     XDebug(DebugAll,"Regexp::Regexp() [%p]",this);
 }
 
 Regexp::Regexp(const char* value, bool extended, bool insensitive)
-    : String(value), m_regexp(0), m_flags(0)
+    : String(value), m_regexp(0), m_compile(true), m_flags(0)
 {
     XDebug(DebugAll,"Regexp::Regexp(\"%s\",%d,%d) [%p]",
 	value,extended,insensitive,this);
@@ -1696,7 +1895,7 @@ Regexp::Regexp(const char* value, bool extended, bool insensitive)
 }
 
 Regexp::Regexp(const Regexp& value)
-    : String(value.c_str()), m_regexp(0), m_flags(value.m_flags)
+    : String(value.c_str()), m_regexp(0), m_compile(true), m_flags(value.m_flags)
 {
     XDebug(DebugAll,"Regexp::Regexp(%p) [%p]",&value,this);
 }
@@ -1729,9 +1928,10 @@ void Regexp::changed()
     String::changed();
 }
 
-bool Regexp::compile() const
+bool Regexp::doCompile() const
 {
     XDebug(DebugInfo,"Regexp::compile()");
+    m_compile = false;
     if (c_str() && !m_regexp) {
 	regex_t *data = (regex_t *) ::malloc(sizeof(regex_t));
 	if (!data) {
@@ -1758,6 +1958,7 @@ void Regexp::cleanup()
 	::regfree(data);
 	::free(data);
     }
+    m_compile = true;
 }
 
 void Regexp::setFlags(bool extended, bool insensitive)
