@@ -101,16 +101,18 @@ namespace TelEngine {
 
 #define DebugMin DebugFail
 #define DebugVis DebugConf
+#define DebugDef DebugNote
 #define DebugMax DebugAll
 
 #define OUT_BUFFER_SIZE 8192
+#define OUT_HEADER_SIZE 112
 
 // RefObject mutex pool array size
 #ifndef REFOBJECT_MUTEX_COUNT
 #define REFOBJECT_MUTEX_COUNT 47
 #endif
 
-static int s_debug = DebugWarn;
+static int s_debug = DebugDef;
 static int s_indent = 0;
 static bool s_debugging = true;
 static bool s_abort = false;
@@ -121,13 +123,13 @@ static Debugger::Formatting s_fmtstamp = Debugger::None;
 static const char* const s_colors[11] = {
     "\033[5;41;1;33m\033[K",// DebugFail - blinking yellow on red
     "\033[44;1;37m\033[K",  // DebugTest - white on blue
-    "\033[41;1;37m\033[K",  // DebugGoOn - white on red
+    "\033[41;1;37m\033[K",  // DebugCrit - white on red
     "\033[41;37m\033[K",    // DebugConf - gray on red
     "\033[40;31m\033[K",    // DebugStub - red on black
     "\033[40;1;31m\033[K",  // DebugWarn - light red on black
     "\033[40;1;33m\033[K",  // DebugMild - yellow on black
-    "\033[40;1;37m\033[K",  // DebugCall - white on black
     "\033[40;1;32m\033[K",  // DebugNote - light green on black
+    "\033[40;1;37m\033[K",  // DebugCall - white on black
     "\033[40;1;36m\033[K",  // DebugInfo - light cyan on black
     "\033[40;36m\033[K"     // DebugAll  - cyan on black
 };
@@ -135,13 +137,13 @@ static const char* const s_colors[11] = {
 static const char* const s_levels[11] = {
     "FAIL",
     "TEST",
-    "GOON",
+    "CRIT",
     "CONF",
     "STUB",
     "WARN",
     "MILD",
-    "CALL",
     "NOTE",
+    "CALL",
     "INFO",
     "ALL",
 };
@@ -306,7 +308,7 @@ void Debug(const char* facility, int level, const char* format, ...)
 	return;
     if (!format)
 	format = "";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     ::snprintf(buf,sizeof(buf),"<%s:%s> ",facility,dbg_level(level));
     va_list va;
     va_start(va,format);
@@ -336,7 +338,7 @@ void Debug(const DebugEnabler* local, int level, const char* format, ...)
 	return;
     if (!format)
 	format = "";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     if (facility)
 	::snprintf(buf,sizeof(buf),"<%s:%s> ",facility,dbg_level(level));
     else
@@ -357,7 +359,7 @@ void Alarm(const char* component, int level, const char* format, ...)
 	return;
     if (TelEngine::null(component))
 	component = "unknown";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     ::snprintf(buf,sizeof(buf),"<%s:%s> ",component,dbg_level(level));
     va_list va;
     va_start(va,format);
@@ -375,7 +377,7 @@ void Alarm(const DebugEnabler* component, int level, const char* format, ...)
 	return;
     const char* name = (component && !TelEngine::null(component->debugName()))
 	? component->debugName() : "unknown";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     ::snprintf(buf,sizeof(buf),"<%s:%s> ",name,dbg_level(level));
     va_list va;
     va_start(va,format);
@@ -393,7 +395,7 @@ void Alarm(const char* component, const char* info, int level, const char* forma
 	return;
     if (TelEngine::null(component))
 	component = "unknown";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     ::snprintf(buf,sizeof(buf),"<%s:%s> ",component,dbg_level(level));
     va_list va;
     va_start(va,format);
@@ -411,7 +413,7 @@ void Alarm(const DebugEnabler* component, const char* info, int level, const cha
 	return;
     const char* name = (component && !TelEngine::null(component->debugName()))
 	? component->debugName() : "unknown";
-    char buf[64];
+    char buf[OUT_HEADER_SIZE];
     ::snprintf(buf,sizeof(buf),"<%s:%s> ",name,dbg_level(level));
     va_list va;
     va_start(va,format);
@@ -504,7 +506,7 @@ Debugger::Debugger(const char* name, const char* format, ...)
     : m_name(name), m_level(DebugAll)
 {
     if (s_debugging && m_name && (s_debug >= DebugAll) && !reentered()) {
-	char buf[64];
+	char buf[OUT_HEADER_SIZE];
 	::snprintf(buf,sizeof(buf),">>> %s",m_name);
 	va_list va;
 	va_start(va,format);
@@ -522,7 +524,7 @@ Debugger::Debugger(int level, const char* name, const char* format, ...)
     : m_name(name), m_level(level)
 {
     if (s_debugging && m_name && (s_debug >= level) && !reentered()) {
-	char buf[64];
+	char buf[OUT_HEADER_SIZE];
 	::snprintf(buf,sizeof(buf),">>> %s",m_name);
 	va_list va;
 	va_start(va,format);
