@@ -34,6 +34,10 @@ public:
 	: JsObject("Object",mtx,true)
 	{
 	}
+    virtual void initConstructor(JsFunction* construct)
+	{
+	    construct->params().addParam(new ExpFunction("keys"));
+	}
 protected:
     bool runNative(ObjList& stack, const ExpOperation& oper, GenObject* context);
 };
@@ -70,6 +74,7 @@ public:
     virtual void initConstructor(JsFunction* construct)
 	{
 	    construct->params().addParam(new ExpFunction("now"));
+	    construct->params().addParam(new ExpFunction("UTC"));
 	}
     virtual JsObject* runConstructor(ObjList& stack, const ExpOperation& oper, GenObject* context);
 protected:
@@ -240,6 +245,148 @@ void JsObject::printRecursive(const GenObject* obj)
     Output("%s",buf.c_str());
 }
 
+String JsObject::strEscape(const char* str)
+{
+    String s("\"");
+    char c;
+    while (str && (c = *str++)) {
+	switch (c) {
+	    case '\"':
+	    case '\\':
+		s += "\\";
+		break;
+	    case '\b':
+		s += "\\b";
+		continue;
+	    case '\f':
+		s += "\\f";
+		continue;
+	    case '\n':
+		s += "\\n";
+		continue;
+	    case '\r':
+		s += "\\r";
+		continue;
+	    case '\t':
+		s += "\\t";
+		continue;
+	    case '\v':
+		s += "\\v";
+		continue;
+	}
+	s += c;
+    }
+    s += "\"";
+    return s;
+}
+
+ExpOperation* JsObject::toJSON(const ExpOperation* oper, int spaces)
+{
+    if (!oper || YOBJECT(JsFunction,oper) || YOBJECT(ExpFunction,oper) || JsParser::isUndefined(*oper))
+	return 0;
+    if (spaces < 0)
+	spaces = 0;
+    else if (spaces > 10)
+	spaces = 10;
+    ExpOperation* ret = new ExpOperation("","JSON");
+    toJSON(oper,*ret,spaces);
+    return ret;
+}
+
+void JsObject::toJSON(const NamedString* ns, String& buf, int spaces, int indent)
+{
+    const ExpOperation* oper = YOBJECT(ExpOperation,ns);
+    if (!oper) {
+	if (ns)
+	    buf << strEscape(*ns);
+	else
+	    buf << "null";
+	return;
+    }
+    if (JsParser::isNull(*oper) || JsParser::isUndefined(*oper) || YOBJECT(JsFunction,oper)
+	    || YOBJECT(ExpFunction,oper)) {
+	buf << "null";
+	return;
+    }
+    const char* nl = spaces ? "\r\n" : "";
+    JsObject* jso = YOBJECT(JsObject,oper);
+    JsArray* jsa = YOBJECT(JsArray,jso);
+    if (jsa) {
+	if (jsa->length() <= 0) {
+	    buf << "[]";
+	    return;
+	}
+	String li(' ',indent);
+	String ci(' ',indent + spaces);
+	buf << "[" << nl;
+	for (int32_t i = 0; ; ) {
+	    buf << ci;
+	    const NamedString* p = jsa->params().getParam(String(i));
+	    if (p)
+		toJSON(p,buf,spaces,indent + spaces);
+	    else
+		buf << "null";
+	    if (++i < jsa->length())
+		buf << "," << nl;
+	    else {
+		buf << nl;
+		break;
+	    }
+	}
+	buf << li << "]";
+	return;
+    }
+    if (jso) {
+	switch (jso->params().count()) {
+	    case 1:
+		if (!jso->params().getParam(protoName()))
+		    break;
+		// fall through
+	    case 0:
+		buf << "{}";
+		return;
+	}
+	ObjList* l = jso->params().paramList()->skipNull();
+	String li(' ',indent);
+	String ci(' ',indent + spaces);
+	const char* sep = spaces ? ": " : ":";
+	buf << "{" << nl;
+	while (l) {
+	    const NamedString* p = static_cast<const NamedString*>(l->get());
+	    l = l->skipNext();
+	    if (p->name() == protoName() || YOBJECT(JsFunction,p) || YOBJECT(ExpFunction,p))
+		continue;
+	    const ExpOperation* op = YOBJECT(ExpOperation,p);
+	    if (op && JsParser::isUndefined(*op))
+		continue;
+	    buf << ci << strEscape(p->name()) << sep;
+	    toJSON(p,buf,spaces,indent + spaces);
+	    for (; l; l = l->skipNext()) {
+		p = static_cast<const NamedString*>(l->get());
+		op = YOBJECT(ExpOperation,p);
+		if (!(p->name() == protoName() || YOBJECT(JsFunction,p) || YOBJECT(ExpFunction,p)
+			|| (op && JsParser::isUndefined(*op))))
+		    break;
+	    }
+	    if (l)
+		buf << ",";
+	    buf << nl;
+	}
+	buf << li << "}";
+	return;
+    }
+    if (oper->isBoolean())
+	buf << String::boolText(oper->valBoolean());
+    else if (oper->isNumber()) {
+	if (oper->isInteger())
+	    buf << oper->number();
+	else
+	    buf << "null";
+    }
+    else
+	buf << strEscape(*oper);
+}
+
 void JsObject::setPrototype(GenObject* context, const String& objName)
 {
     ScriptContext* ctxt = YOBJECT(ScriptContext,context);
@@ -266,7 +413,7 @@ JsObject* JsObject::buildCallContext(Mutex* mtx, JsObject* thisObj)
 
 void JsObject::fillFieldNames(ObjList& names)
 {
-    ScriptContext::fillFieldNames(names,params(),"__");
+    ScriptContext::fillFieldNames(names,params(),false,"__");
     const NamedList* native = nativeParams();
     if (native)
 	ScriptContext::fillFieldNames(names,*native);
@@ -449,6 +596,9 @@ int JsObject::extractArgs(JsObject* obj, ObjList& stack, const ExpOperation& ope
 	return 0;
     for (int i = (int)oper.number(); i;  i--) {
 	ExpOperation* op = obj->popValue(stack,context);
+	JsFunction* jsf = YOBJECT(JsFunction,op);
+	if (jsf)
+	    jsf->firstName(op->name());
 	arguments.insert(op);
     }
     return (int)oper.number();
@@ -495,6 +645,37 @@ bool JsObjectObj::runNative(ObjList& stack, const ExpOperation& oper, GenObject*
 {
     if (oper.name() == YSTRING("constructor"))
 	ExpEvaluator::pushOne(stack,new ExpWrapper(new JsObject("Object",mutex())));
+    else if (oper.name() == YSTRING("keys")) {
+	ExpOperation* op = 0;
+	GenObject* obj = 0;
+	if (oper.number() == 0) {
+	    ScriptRun* run = YOBJECT(ScriptRun,context);
+	    if (run)
+		obj = run->context();
+	    else
+		obj = context;
+	}
+	else if (oper.number() == 1) {
+	    op = popValue(stack,context);
+	    if (!op)
+		return false;
+	    obj = op;
+	}
+	else
+	    return false;
+	const NamedList* lst = YOBJECT(NamedList,obj);
+	if (lst) {
+	    NamedIterator iter(*lst);
+	    JsArray* jsa = new JsArray(context,mutex());
+	    while (const NamedString* ns = iter.get())
+		if (ns->name() != protoName())
+		    jsa->push(new ExpOperation(ns->name(),0,true));
+	    ExpEvaluator::pushOne(stack,new ExpWrapper(jsa,"keys"));
+	}
+	else
+	    ExpEvaluator::pushOne(stack,JsParser::nullClone());
+	TelEngine::destruct(op);
+    }
     else
 	return JsObject::runNative(stack,oper,context);
     return true;
@@ -514,7 +695,9 @@ JsArray::JsArray(Mutex* mtx)
     params().addParam(new ExpFunction("slice"));
     params().addParam(new ExpFunction("splice"));
     params().addParam(new ExpFunction("sort"));
+    params().addParam(new ExpFunction("includes"));
     params().addParam(new ExpFunction("indexOf"));
+    params().addParam(new ExpFunction("lastIndexOf"));
     params().addParam("length","0");
 }
 
@@ -587,11 +770,11 @@ JsObject* JsArray::runConstructor(ObjList& stack, const ExpOperation& oper, GenO
     if (!ref())
 	return 0;
     JsArray* obj = static_cast<JsArray*>(clone("[object " + oper.name() + "]"));
-    unsigned int len = oper.number();
+    unsigned int len = (unsigned int)oper.number();
     for (unsigned int i = len; i;  i--) {
 	ExpOperation* op = obj->popValue(stack,context);
 	if ((len == 1) && op->isInteger() && (op->number() >= 0) && (op->number() <= 0xffffffff)) {
-	    len = op->number();
+	    len = (unsigned int)op->number();
 	    TelEngine::destruct(op);
 	    break;
 	}
@@ -611,7 +794,7 @@ bool JsArray::runNative(ObjList& stack, const ExpOperation& oper, GenObject* con
 	// Static function that checks if the argument is an Array
 	ObjList args;
 	extractArgs(this,stack,oper,context,args);
-	ExpEvaluator::pushOne(stack,new ExpOperation(YOBJECT(JsArray,args[0])));
+	ExpEvaluator::pushOne(stack,new ExpOperation(!!YOBJECT(JsArray,args[0])));
     }
     else if (oper.name() == YSTRING("push")) {
 	// Adds one or more elements to the end of an array and returns the new length of the array.
@@ -777,22 +960,24 @@ bool JsArray::runNative(ObjList& stack, const ExpOperation& oper, GenObject* con
 	// New length: 4
 	// shift array
 	int32_t shift = (int32_t)oper.number();
-	for (int32_t i = length() + shift - 1; i >= shift; i--) {
-	    NamedString* ns = static_cast<NamedString*>((*params().paramList())[String(i - shift)]);
-	    if (ns) {
-		String index(i);
-		params().clearParam(index);
-		const_cast<String&>(ns->name()) = index;
+	if (shift >= 1) {
+	    for (int32_t i = length() + shift - 1; i >= shift; i--) {
+		NamedString* ns = static_cast<NamedString*>((*params().paramList())[String(i - shift)]);
+		if (ns) {
+		    String index(i);
+		    params().clearParam(index);
+		    const_cast<String&>(ns->name()) = index;
+		}
 	    }
+	    for (int32_t i = shift - 1; i >= 0; i--) {
+		ExpOperation* op = popValue(stack,context);
+		if (!op)
+		    continue;
+	        const_cast<String&>(op->name()) = i;
+		params().paramList()->insert(op);
+	    }
+	    setLength(length() + shift);
 	}
-	for (int32_t i = shift - 1; i >= 0; i--) {
-	    ExpOperation* op = popValue(stack,context);
-	    if (!op)
-		continue;
-	    const_cast<String&>(op->name()) = i;
-	    params().paramList()->insert(op);
-	}
-	setLength(length() + shift);
 	ExpEvaluator::pushOne(stack,new ExpOperation((int64_t)length()));
     }
     else if (oper.name() == YSTRING("slice"))
@@ -811,7 +996,11 @@ bool JsArray::runNative(ObjList& stack, const ExpOperation& oper, GenObject* con
 	for (int32_t i = 0; i < length(); i++)
 	    result.append(params()[String(i)],separator);
 	ExpEvaluator::pushOne(stack,new ExpOperation(result));
-    } else if (oper.name() == YSTRING("indexOf")) {
+    } else if (oper.name() == YSTRING("includes") || oper.name() == YSTRING("indexOf")
+	    || oper.name() == YSTRING("lastIndexOf")) {
+	// arr.includes(searchElement[,startIndex = 0[,"fieldName"]])
+	// arr.indexOf(searchElement[,startIndex = 0[,"fieldName"]])
+	// arr.lastIndexOf(searchElement[,startIndex = arr.length-1[,"fieldName"]])
 	ObjList args;
 	if (!extractArgs(this,stack,oper,context,args)) {
 	    Debug(DebugWarn,"Failed to extract arguments!");
@@ -820,19 +1009,47 @@ bool JsArray::runNative(ObjList& stack, const ExpOperation& oper, GenObject* con
 	ExpOperation* op1 = static_cast<ExpOperation*>(args.remove(false));
 	if (!op1)
 	    return false;
+	ExpWrapper* w1 = YOBJECT(ExpWrapper,op1);
+	ExpOperation* fld = 0;
+	int dir = 1;
 	int pos = 0;
+	if (oper.name().at(0) == 'l') {
+	    dir = -1;
+	    pos = length() - 1;
+	}
 	if (args.skipNull()) {
 	    String* spos = static_cast<String*>(args.remove(false));
-	    if (spos)
-		pos = spos->toInteger(0);
+	    if (spos) {
+		pos = spos->toInteger(pos);
+		if (pos < 0)
+		    pos += length();
+		if (dir > 0) {
+		    if (pos < 0)
+			pos = 0;
+		}
+		else if (pos >= length())
+		    pos = length() - 1;
+	    }
 	    TelEngine::destruct(spos);
+	    fld = static_cast<ExpOperation*>(args.remove(false));
 	}
 	int index = -1;
-	for (int i = pos;i < length();i++) {
+	for (int i = pos; ; i += dir) {
+	    if (dir > 0) {
+		if (i >= length())
+		    break;
+	    }
+	    else if (i < 0)
+		break;
 	    ExpOperation* op2 = static_cast<ExpOperation*>(params().getParam(String(i)));
+	    if (op2 && !TelEngine::null(fld)) {
+		const ExpExtender* ext = YOBJECT(ExpExtender,op2);
+		if (!ext)
+		    continue;
+		op2 = YOBJECT(ExpOperation,ext->getField(stack,*fld,context));
+	    }
 	    if (!op2 || op2->opcode() != op1->opcode())
 		continue;
-	    ExpWrapper* w1 = YOBJECT(ExpWrapper,op1);
 	    ExpWrapper* w2 = YOBJECT(ExpWrapper,op2);
 	    if (w1 || w2) {
 		if (w1 && w2 && w1->object() == w2->object()) {
@@ -845,7 +1062,11 @@ bool JsArray::runNative(ObjList& stack, const ExpOperation& oper, GenObject* con
 	    }
 	}
 	TelEngine::destruct(op1);
-	ExpEvaluator::pushOne(stack,new ExpOperation((int64_t)index));
+	TelEngine::destruct(fld);
+	if (oper.name().length() == 8)
+	    ExpEvaluator::pushOne(stack,new ExpOperation(index >= 0));
+	else
+	    ExpEvaluator::pushOne(stack,new ExpOperation((int64_t)index));
 	return true;
     }
     else
@@ -1073,24 +1294,32 @@ JsRegExp::JsRegExp(Mutex* mtx)
     : JsObject("RegExp",mtx)
 {
     params().addParam(new ExpFunction("test"));
+    params().addParam(new ExpFunction("valid"));
 }
 
 JsRegExp::JsRegExp(Mutex* mtx, const char* name, const char* rexp, bool insensitive, bool extended, bool frozen)
     : JsObject(mtx,name,frozen),
       m_regexp(rexp,extended,insensitive)
 {
-    params().addParam(new ExpFunction("test"));
+    XDebug(DebugAll,"JsRegExp::JsRegExp('%s',%p,%s) [%p]",
+	name,mtx,String::boolText(frozen),this);
     params().addParam("ignoreCase",String::boolText(insensitive));
     params().addParam("basicPosix",String::boolText(!extended));
 }
 
 JsRegExp::JsRegExp(Mutex* mtx, const Regexp& rexp, bool frozen)
-    : JsObject("RegExp",mtx),
+    : JsObject(mtx,rexp.c_str()),
       m_regexp(rexp)
 {
-    params().addParam(new ExpFunction("test"));
-    params().addParam("ignoreCase",String::boolText(rexp.isCaseInsensitive()));
-    params().addParam("basicPosix",String::boolText(!rexp.isExtended()));
+    XDebug(DebugAll,"JsRegExp::JsRegExp('%s',%p,%s) [%p]",
+	toString().c_str(),mtx,String::boolText(frozen),this);
+}
+
+JsObject* JsRegExp::copy(Mutex* mtx) const
+{
+    JsRegExp* reg = new JsRegExp(mtx,m_regexp,frozen());
+    deepCopyParams(reg->params(),params(),mtx);
+    return reg;
 }
 
 bool JsRegExp::runNative(ObjList& stack, const ExpOperation& oper, GenObject* context)
@@ -1105,11 +1334,29 @@ bool JsRegExp::runNative(ObjList& stack, const ExpOperation& oper, GenObject* co
 	TelEngine::destruct(op);
 	ExpEvaluator::pushOne(stack,new ExpOperation(ok));
     }
+    else if (oper.name() == YSTRING("valid")) {
+	if (oper.number())
+	    return false;
+	ExpEvaluator::pushOne(stack,new ExpOperation(regexp().compile()));
+    }
     else
 	return JsObject::runNative(stack,oper,context);
     return true;
 }
 
+bool JsRegExp::runAssign(ObjList& stack, const ExpOperation& oper, GenObject* context)
+{
+    XDebug(DebugAll,"JsRegExp::runAssign() '%s'='%s' (%s) in '%s' [%p]",
+	oper.name().c_str(),oper.c_str(),oper.typeOf(),toString().c_str(),this);
+    if (!JsObject::runAssign(stack,oper,context))
+	return false;
+    if (oper.name() == YSTRING("ignoreCase"))
+	regexp().setFlags(regexp().isExtended(),oper.toBoolean());
+    else if (oper.name() == YSTRING("basicPosix"))
+	regexp().setFlags(!oper.toBoolean(),regexp().isCaseInsensitive());
+    return true;
+}
+
 JsObject* JsRegExp::runConstructor(ObjList& stack, const ExpOperation& oper, GenObject* context)
 {
     ObjList args;
@@ -1126,7 +1373,7 @@ JsObject* JsRegExp::runConstructor(ObjList& stack, const ExpOperation& oper, Gen
 	return 0;
     bool insensitive = false;
     bool extended = true;
-    if (flags)  {
+    if (flags && *flags)  {
 	const char* f = *flags;
 	char c = *f++;
 	while (c) {
@@ -1212,11 +1459,8 @@ bool JsMath::runNative(ObjList& stack, const ExpOperation& oper, GenObject* cont
 	}
 	if (min < 0 || max < 0 || min >= max)
 	    return false;
-	unsigned long interval = max;
-	if (min != 0)
-	    interval -= min + 1;
-	int64_t rand = (Random::random() % interval) + min;
-	ExpEvaluator::pushOne(stack,new ExpOperation(rand));
+	int64_t rand = (max > (min + 1)) ? (Random::random() % (max - min)) : 0;
+	ExpEvaluator::pushOne(stack,new ExpOperation(rand + min));
     }
     else
 	return JsObject::runNative(stack,oper,context);
@@ -1226,7 +1470,7 @@ bool JsMath::runNative(ObjList& stack, const ExpOperation& oper, GenObject* cont
 
 JsObject* JsDate::runConstructor(ObjList& stack, const ExpOperation& oper, GenObject* context)
 {
-    XDebug(DebugAll,"JsDate::runConstructor '%s'("FMT64")",oper.name().c_str(),oper.number());
+    XDebug(DebugAll,"JsDate::runConstructor '%s'(" FMT64 ")",oper.name().c_str(),oper.number());
     ObjList args;
     JsObject* obj = 0;
     switch (extractArgs(stack,oper,context,args)) {
@@ -1240,13 +1484,16 @@ JsObject* JsDate::runConstructor(ObjList& stack, const ExpOperation& oper, GenOb
 		    obj = new JsDate(mutex(),val->number());
 	    }
 	    break;
+	case 2:
 	case 3:
+	case 4:
+	case 5:
 	case 6:
 	case 7:
 	    {
 		unsigned int parts[7];
 		for (int i = 0; i < 7; i++) {
-		    parts[i] = 0;
+		    parts[i] = (2 == i) ? 1 : 0;
 		    ExpOperation* val = static_cast<ExpOperation*>(args[i]);
 		    if (val) {
 			if (val->isInteger())
@@ -1255,9 +1502,10 @@ JsObject* JsDate::runConstructor(ObjList& stack, const ExpOperation& oper, GenOb
 			    return 0;
 		    }
 		}
-		// Date components use local time, month starts from 0
-		if (parts[1] < 12)
-		    parts[1]++;
+		// Date components use local time, year can be 0-99, month starts from 0
+		if (parts[0] < 100)
+		    parts[0] += 1900;
+		parts[1]++;
 		u_int64_t time = Time::toEpoch(parts[0],parts[1],parts[2],parts[3],parts[4],parts[5]);
 		obj = new JsDate(mutex(),1000 * time + parts[6],true);
 	    }
@@ -1278,6 +1526,46 @@ bool JsDate::runNative(ObjList& stack, const ExpOperation& oper, GenObject* cont
 	// Returns the number of milliseconds elapsed since 1 January 1970 00:00:00 UTC.
 	ExpEvaluator::pushOne(stack,new ExpOperation((int64_t)Time::msecNow()));
     }
+    else if (oper.name() == YSTRING("UTC")) {
+	ObjList args;
+	switch (extractArgs(stack,oper,context,args)) {
+	    case 2:
+	    case 3:
+	    case 4:
+	    case 5:
+	    case 6:
+	    case 7:
+		{
+		    unsigned int parts[7];
+		    for (int i = 0; i < 7; i++) {
+			parts[i] = (2 == i) ? 1 : 0;
+			ExpOperation* val = static_cast<ExpOperation*>(args[i]);
+			if (val) {
+			    if (val->isInteger())
+				parts[i] = (int)val->number();
+			    else
+				return 0;
+			}
+		    }
+		    // Year can be 0-99, month starts from 0
+		    if (parts[0] < 100)
+			parts[0] += 1900;
+		    parts[1]++;
+		    unsigned int time = Time::toEpoch(parts[0],parts[1],parts[2],parts[3],parts[4],parts[5]);
+		    if (time != (unsigned int)-1) {
+			ExpEvaluator::pushOne(stack,new ExpOperation(1000 * (int64_t)time + parts[6]));
+			break;
+		    }
+		}
+		// fall through
+	    case 0:
+	    case 1:
+		ExpEvaluator::pushOne(stack,new ExpOperation(ExpOperation::nonInteger(),"NaN"));
+		break;
+	    default:
+		return false;
+	}
+    }
     else if (oper.name() == YSTRING("getDate")) {
 	// Returns the day of the month for the specified date according to local time.
 	// The value returned by getDate is an integer between 1 and 31.
