@@ -55,42 +55,71 @@ const TokenDict SDPParser::s_payloads[] = {
     { "mjpeg",        26 },
     { "h261",         31 },
     { "h263",         34 },
+    { "h263-1998",   111 },
+    { "h263-2000",   112 },
+    { "h264",        114 },
+    { "vp8",         113 },
     { "mpv",          32 },
     { "mp2t",         33 },
-    { "mp4v",         98 },
-    {      0,          0 },
+    { "mp4v",        110 },
+    { 0,               0 },
 };
 
 // SDP Payloads for the AV profile
 const TokenDict SDPParser::s_rtpmap[] = {
-    { "PCMU/8000",     0 },
-    { "PCMA/8000",     8 },
-    { "GSM/8000",      3 },
-    { "LPC/8000",      7 },
-    { "L16/8000",     11 },
-    { "G726-32/8000",  2 },
-    { "G722/8000",     9 },
-    { "G723/8000",     4 },
-    { "G728/8000",    15 },
-    { "G729/8000",    18 },
-    { "G729A/8000",   18 },
-    { "MPA/90000",    14 },
-    { "iLBC/8000",    98 },
-    { "AMR/8000",     96 },
-    { "AMR-WB/16000", 99 },
-    { "SPEEX/8000",  102 },
-    { "SPEEX/16000", 103 },
-    { "SPEEX/32000", 104 },
-    { "iSAC/16000",  105 },
-    { "iSAC/32000",  106 },
-    { "GSM-EFR/8000",107 },
-    { "JPEG/90000",   26 },
-    { "H261/90000",   31 },
-    { "H263/90000",   34 },
-    { "MPV/90000",    32 },
-    { "MP2T/90000",   33 },
-    { "MP4V-ES/90000",98 },
-    {           0,     0 },
+    { "PCMU/8000",          0 },
+    { "PCMA/8000",          8 },
+    { "GSM/8000",           3 },
+    { "LPC/8000",           7 },
+    { "L16/8000",          11 },
+    { "G726-32/8000",       2 },
+    { "G722/8000",          9 },
+    { "G723/8000",          4 },
+    { "G728/8000",         15 },
+    { "G729/8000",         18 },
+    { "G729A/8000",        18 },
+    { "MPA/90000",         14 },
+    { "iLBC/8000",         98 },
+    { "AMR/8000",          96 },
+    { "AMR-WB/16000",      99 },
+    { "SPEEX/8000",       102 },
+    { "SPEEX/16000",      103 },
+    { "SPEEX/32000",      104 },
+    { "iSAC/16000",       105 },
+    { "iSAC/32000",       106 },
+    { "GSM-EFR/8000",     107 },
+    { "JPEG/90000",        26 },
+    { "H261/90000",        31 },
+    { "H263/90000",        34 },
+    { "H263-1998/90000",  111 },
+    { "H263-2000/90000",  112 },
+    { "H264/90000",       114 },
+    { "VP8/90000",        113 },
+    { "MPV/90000",         32 },
+    { "MP2T/90000",        33 },
+    { "MP4V-ES/90000",    110 },
+    { 0,                    0 },
+};
+
+enum SdpFormat
+{
+    SdpFmtUnknown = 0,
+    SdpIlbc,
+    SdpG729,
+    SdpAmr
+};
+
+// We need to check parameters (fmtp line) for these formats to detect variants
+static const TokenDict s_sdpFmtParamsCheck[] = {
+    { "g729",         SdpG729 },
+    { "ilbc",         SdpIlbc },
+    { "ilbc20",       SdpIlbc },
+    { "ilbc30",       SdpIlbc },
+    { "amr",          SdpAmr },
+    { "amr-o",        SdpAmr },
+    { "amr/16000",    SdpAmr },
+    { "amr-o/16000",  SdpAmr },
+    { 0,              0 },
 };
 
 // Parse a received SDP body
@@ -154,6 +183,7 @@ ObjList* SDPParser::parse(const MimeSdpBody& sdp, String& addr, ObjList* oldMedi
 	String aux;
 	String mappings;
 	String crypto;
+	ObjList dups;
 	ObjList params;
 	ObjList* dest = &params;
 	bool first = true;
@@ -175,17 +205,34 @@ ObjList* SDPParser::parse(const MimeSdpBody& sdp, String& addr, ObjList* oldMedi
 		fmt = tmp;
 		tmp.clear();
 	    }
+	    ObjList fmtl;
 	    int mode = 0;
 	    bool annexB = m_codecs.getBoolValue("g729_annexb",false);
 	    bool amrOctet = m_codecs.getBoolValue("amr_octet",false);
 	    int defmap = -1;
 	    String payload(lookup(var,s_payloads));
+	    if (amrOctet) {
+		if (payload == YSTRING("amr"))
+		    payload = "amr-o";
+		else if (payload == YSTRING("amr/16000"))
+		    payload = "amr-o/16000";
+	    }
 
 	    const ObjList* l = sdp.lines().find(c);
 	    while (l && (l = l->skipNext())) {
 		const NamedString* s = static_cast<NamedString*>(l->get());
 		if (s->name() == "m")
 		    break;
+		if (s->name() == "b") {
+		    if (first) {
+			int pos = s->find(':');
+			if (pos >= 0)
+			    dest = dest->append(new NamedString("BW-" + s->substr(0,pos),s->substr(pos+1)));
+			else
+			    dest = dest->append(new NamedString("BW-" + *s));
+		    }
+		    continue;
+		}
 		if (s->name() != "a")
 		    continue;
 		String line(*s);
@@ -218,15 +265,77 @@ ObjList* SDPParser::parse(const MimeSdpBody& sdp, String& addr, ObjList* oldMedi
 		    }
 		}
 		else if (line.startSkip("fmtp:",false)) {
+		    if (payload.null())
+			continue;
 		    int num = var - 1;
-		    line >> num >> " ";
-		    if (num == var) {
-			if (line.startSkip("mode=",false))
-			    line >> mode;
-			else if (line.startSkip("annexb=",false))
-			    line >> annexB;
-			else if (line.startSkip("octet-align=",false))
-			    amrOctet = (0 != line.toInteger(0));
+		    line >> num;
+		    if ((num == var) && line.trimSpaces()) {
+			bool found = false;
+			ObjList* fmtps = 0;
+			int checkParams = lookup(payload,s_sdpFmtParamsCheck);
+			if (checkParams) {
+			    fmtps = line.split(';',false);
+			    for (ObjList* o = fmtps->skipNull(); o; o = o->skipNext()) {
+				String& s = *static_cast<String*>(o->get());
+				switch (checkParams) {
+				    case SdpIlbc:
+					if (!s.startSkip("mode=",false))
+					    continue;
+					s >> mode;
+					break;
+				    case SdpG729:
+					if (!s.startSkip("annexb=",false))
+					    continue;
+					s >> annexB;
+					break;
+				    case SdpAmr:
+					if (s.startSkip("octet-align=",false)) {
+					    int val = 0;
+					    s >> val;
+					    amrOctet = (0 != val);
+					    if (amrOctet) {
+						if (payload == YSTRING("amr"))
+						    payload = "amr-o";
+						else if (payload == YSTRING("amr/16000"))
+						    payload = "amr-o/16000";
+					    }
+					    else if (payload == YSTRING("amr-o"))
+						payload = "amr";
+					    else if (payload == YSTRING("amr-o/16000"))
+						payload = "amr/16000";
+					    break;
+					}
+					continue;
+				    default:
+					continue;
+				}
+				found = true;
+				o->set(0);
+				// Done: we are searching for a single parameter
+				break;
+			    }
+			}
+			XDebug(this,DebugAll,"%s fmtp '%s' (%d) '%s'",
+			    (checkParams ? (found ? "Found" : "Checked") : "Parsed"),
+			    payload.c_str(),var,line.c_str());
+			if (!found)
+			    fmtl.append(new NamedString("fmtp:" + payload,line));
+			else if (fmtps && fmtps->skipNull()) {
+			    line.clear();
+			    fmtl.append(new NamedString("fmtp:" + payload,line.append(fmtps,";")));
+			}
+			TelEngine::destruct(fmtps);
+		    }
+		}
+		else if (line.startSkip("gpmd:",false)) {
+		    if (payload.null())
+			continue;
+		    int num = var - 1;
+		    line >> num;
+		    if ((num == var) && line.trimSpaces()) {
+			XDebug(this,DebugAll,"Found gpmd '%s' (%d) '%s'",
+			    payload.c_str(),var,line.c_str());
+			fmtl.append(new NamedString("gpmd:" + payload,line));
 		    }
 		}
 		else if (first) {
@@ -265,11 +374,12 @@ ObjList* SDPParser::parse(const MimeSdpBody& sdp, String& addr, ObjList* oldMedi
 		    payload = m_hacks.getValue(YSTRING("ilbc_default"),"ilbc30");
 	    }
 
-	    if (amrOctet && payload == "amr")
-		payload = "amr-o";
-
-	    XDebug(this,DebugAll,"Payload %d format '%s'",var,payload.c_str());
-	    if (payload && m_codecs.getBoolValue(payload,defcodecs && DataTranslator::canConvert(payload))) {
+	    XDebug(this,DebugAll,"Payload %d format '%s'%s",var,payload.c_str(),
+		(dups.find(payload) ? " (duplicated)" : ""));
+	    if (payload.null() || dups.find(payload))
+		continue;
+	    dups.append(new String(payload));
+	    if (m_codecs.getBoolValue(payload,defcodecs && DataTranslator::canConvert(payload))) {
 		if (fmt)
 		    fmt << ",";
 		fmt << payload;
@@ -278,6 +388,8 @@ ObjList* SDPParser::parse(const MimeSdpBody& sdp, String& addr, ObjList* oldMedi
 			mappings << ",";
 		    mappings << payload << "=" << var;
 		}
+		while (NamedString* par = static_cast<NamedString*>(fmtl.remove(false)))
+		    dest = dest->append(par);
 		if ((payload == "g729") && m_hacks.getBoolValue(YSTRING("g729_annexb"),annexB))
 		    aux << ",g729b";
 	    }
