@@ -31,7 +31,7 @@ SDPSession::SDPSession(SDPParser* parser)
       m_rtpForward(false), m_sdpForward(false), m_rtpMedia(0),
       m_sdpSession(0), m_sdpVersion(0), m_sdpHash(YSTRING_INIT_HASH),
       m_secure(m_parser->m_secure), m_rfc2833(m_parser->m_rfc2833),
-      m_ipv6(false), m_enabler(0), m_ptr(0)
+      m_ipv6(false), m_amrExtra(""), m_enabler(0), m_ptr(0)
 {
     setSdpDebug();
 }
@@ -40,7 +40,7 @@ SDPSession::SDPSession(SDPParser* parser, NamedList& params)
     : m_parser(parser), m_mediaStatus(MediaMissing),
       m_rtpForward(false), m_sdpForward(false), m_rtpMedia(0),
       m_sdpSession(0), m_sdpVersion(0), m_sdpHash(YSTRING_INIT_HASH),
-      m_ipv6(false), m_enabler(0), m_ptr(0)
+      m_ipv6(false), m_amrExtra(""), m_enabler(0), m_ptr(0)
 {
     setSdpDebug();
     m_rtpForward = params.getBoolValue("rtp_forward");
@@ -130,15 +130,17 @@ bool SDPSession::dispatchRtp(SDPMedia* media, const char* addr, bool start,
     m_rtpForward = false;
     m_rtpLocalAddr = m->getValue("localip",m_rtpLocalAddr);
     m_mediaStatus = m_rtpLocalAddr.null() ? MediaMuted : MediaStarted;
-    const char* sdpPrefix = m->getValue("osdp-prefix","osdp");
+    String sdpPrefix = m->getValue("osdp-prefix","osdp");
     if (sdpPrefix) {
+	if (!sdpPrefix.endsWith("_"))
+	    sdpPrefix += "_";
 	unsigned int n = m->length();
 	for (unsigned int j = 0; j < n; j++) {
 	    const NamedString* param = m->getParam(j);
 	    if (!param)
 		continue;
 	    String tmp = param->name();
-	    if (tmp.startSkip(sdpPrefix,false) && tmp.startSkip("_",false) && tmp)
+	    if (tmp.startSkip(sdpPrefix,false) && tmp)
 	        media->parameter(tmp,*param,false);
 	}
     }
@@ -202,10 +204,10 @@ bool SDPSession::startRtp(RefObject* context)
 }
 
 // Update from parameters. Build a default SDP if no media is found in params
-bool SDPSession::updateSDP(const NamedList& params)
+bool SDPSession::updateSDP(const NamedList& params, bool defaults)
 {
-    DDebug(m_enabler,DebugAll,"SDPSession::updateSdp('%s') [%p]",params.c_str(),m_ptr);
-    bool defaults = true;
+    DDebug(m_enabler,DebugAll,"SDPSession::updateSdp('%s',%s) [%p]",
+	params.c_str(),String::boolText(defaults),m_ptr);
     const char* sdpPrefix = params.getValue("osdp-prefix","osdp");
     ObjList* lst = 0;
     unsigned int n = params.length();
@@ -215,7 +217,7 @@ bool SDPSession::updateSDP(const NamedList& params)
 	const NamedString* p = params.getParam(i);
 	if (!p)
 	    continue;
-	// search for rtp_port or rtp_port_MEDIANAME parameters
+	// search for media or media_MEDIANAME parameters
 	String tmp(p->name());
 	if (!tmp.startSkip("media",false))
 	    continue;
@@ -257,12 +259,14 @@ bool SDPSession::updateSDP(const NamedList& params)
 	}
 	rtp->crypto(crypto,false);
 	if (sdpPrefix) {
+	    String prefix;
+	    prefix << sdpPrefix << rtp->suffix() << "_";
 	    for (unsigned int j = 0; j < n; j++) {
 		const NamedString* param = params.getParam(j);
 		if (!param)
 		    continue;
 		tmp = param->name();
-		if (tmp.startSkip(sdpPrefix + rtp->suffix() + "_",false) && (tmp.find('_') < 0))
+		if (tmp.startSkip(prefix,false) && (tmp.find('_') < 0))
 		    rtp->parameter(tmp,*param,append);
 	    }
 	}
@@ -318,15 +322,28 @@ static int addIP(String& buf, const char* addr, int family = SocketAddr::Unknown
     return family;
 }
 
+// Utility used in createSDP
+// Build a 'fmtp' line if not already done. Add a parameter to it
+static inline void setFmtpLine(String*& line, int payload, const char* param)
+{
+    if (line)
+	line->append(param,";");
+    else {
+	line = new String("fmtp:");
+	*line << payload << " " << param;
+    }
+}
+
 // Creates a SDP body from transport address and list of media descriptors
 // Use own list if given media list is 0
 MimeSdpBody* SDPSession::createSDP(const char* addr, ObjList* mediaList)
 {
     DDebug(m_enabler,DebugAll,"SDPSession::createSDP('%s',%p) [%p]",addr,mediaList,m_ptr);
-    if (!mediaList)
-	mediaList = m_rtpMedia;
+    ObjList* mList = mediaList;
+    if (!mList)
+	mList = m_rtpMedia;
     // if we got no media descriptors we simply create no SDP
-    if (!mediaList)
+    if (!mList)
 	return 0;
     if (!m_sdpSession)
 	m_sdpVersion = m_sdpSession = Time::secNow();
@@ -352,7 +369,7 @@ MimeSdpBody* SDPSession::createSDP(const char* addr, ObjList* mediaList)
 
     Lock lock(m_parser);
     bool defcodecs = m_parser->m_codecs.getBoolValue("default",true);
-    for (ObjList* ml = mediaList->skipNull(); ml; ml = ml->skipNext()) {
+    for (ObjList* ml = mList->skipNull(); ml; ml = ml->skipNext()) {
 	SDPMedia* m = static_cast<SDPMedia*>(ml->get());
 	int rfc2833 = 0;
 	if ((m_rfc2833 >= 0) && m->isAudio()) {
@@ -380,7 +397,6 @@ MimeSdpBody* SDPSession::createSDP(const char* addr, ObjList* mediaList)
 	for (; f; f = f->next()) {
 	    const String* s = static_cast<const String*>(f->get());
 	    if (s) {
-		int mode = 0;
 		if (*s == "g729b")
 		    continue;
 		int payload = s->toInteger(SDPParser::s_payloads,-1);
@@ -454,34 +470,50 @@ MimeSdpBody* SDPSession::createSDP(const char* addr, ObjList* mediaList)
 			defcode = payload;
 		    const char* map = lookup(defcode,SDPParser::s_rtpmap);
 		    if (map && m_parser->m_codecs.getBoolValue(*s,defcodecs && DataTranslator::canConvert(*s))) {
-			if (*s == "ilbc20")
-			    ptime = mode = 20;
-			else if (*s == "ilbc30")
-			    ptime = mode = 30;
 			frm << " " << payload;
 			String* temp = new String("rtpmap:");
 			*temp << payload << " " << map;
 			dest = dest->append(temp);
-			if (mode) {
-			    temp = new String("fmtp:");
-			    *temp << payload << " mode=" << mode;
-			    dest = dest->append(temp);
+			const String* fmtp = m->getParam("fmtp:" + *s);
+			temp = 0;
+			if (*s == YSTRING("ilbc20")) {
+			    ptime = 20;
+			    setFmtpLine(temp,payload,"mode=20");
 			}
-			if (*s == "g729") {
-			    temp = new String("fmtp:");
-			    *temp << payload << " annexb=" <<
-				((0 != l->find("g729b")) ? "yes" : "no");
-			    dest = dest->append(temp);
+			else if (*s == YSTRING("ilbc30")) {
+			    ptime = 30;
+			    setFmtpLine(temp,payload,"mode=30");
 			}
-			else if (*s == "amr") {
-			    temp = new String("fmtp:");
-			    *temp << payload << " octet-align=0";
-			    dest = dest->append(temp);
+			else if (*s == YSTRING("g729")) {
+			    setFmtpLine(temp,payload,"annexb=");
+			    *temp << ((0 != l->find("g729b")) ? "yes" : "no");
 			}
-			else if (*s == "amr-o") {
-			    temp = new String("fmtp:");
-			    *temp << payload << " octet-align=1";
+			else if (*s == YSTRING("amr") || s->startsWith("amr/")) {
+			    setFmtpLine(temp,payload,"octet-align=0");
+			    if (!m_rtpForward && m_amrExtra) {
+				addFmtpAmrExtra(*temp,fmtp);
+				fmtp = 0;
+			    }
+			}
+			else if (*s == YSTRING("amr-o") || s->startsWith("amr-o/")) {
+			    setFmtpLine(temp,payload,"octet-align=1");
+			    if (!m_rtpForward && m_amrExtra) {
+				addFmtpAmrExtra(*temp,fmtp);
+				fmtp = 0;
+			    }
+			}
+			if (fmtp)
+			    setFmtpLine(temp,payload,*fmtp);
+			if (temp)
 			    dest = dest->append(temp);
+			if (mediaList) {
+			    // RTP forward propagates General Purpose Media Descriptor
+			    const String* gpmd = m->getParam("gpmd:" + *s);
+			    if (gpmd) {
+				temp = new String("gpmd:");
+				*temp << payload << " " << *gpmd;
+				dest = dest->append(temp);
+			    }
 			}
 		    }
 		}
@@ -526,12 +558,19 @@ MimeSdpBody* SDPSession::createSDP(const char* addr, ObjList* mediaList)
 	    unsigned int n = m->length();
 	    for (unsigned int i = 0; i < n; i++) {
 		const NamedString* param = m->getParam(i);
-		if (param) {
+		if (param && (param->name().find(':') < 0)) {
+		    const char* type = "a";
 		    String tmp = param->name();
+		    if (tmp.startSkip("BW-",false)) {
+			if (!tmp)
+			    continue;
+			type = "b";
+		    }
+		    else
+			enc = enc || (tmp == "encryption");
 		    if (*param)
 			tmp << ":" << *param;
-		    sdp->addLine("a",tmp);
-		    enc = enc || (param->name() == "encryption");
+		    sdp->addLine(type,tmp);
 		}
 	    }
 	}
@@ -618,7 +657,7 @@ void SDPSession::updateFormats(const NamedList& msg, bool changeMedia)
 	    const NamedString* p = msg.getParam(i);
 	    if (!p)
 		continue;
-	    // search for media_MEDIANAME parameters
+	    // search for media or media_MEDIANAME parameters
 	    String tmp = p->name();
 	    if (!tmp.startSkip("media",false))
 		continue;
@@ -675,6 +714,33 @@ void SDPSession::updateFormats(const NamedList& msg, bool changeMedia)
 	    }
 	}
     }
+    String sdpPrefix = msg.getValue("osdp-prefix");
+    if (!sdpPrefix)
+	return;
+    if (!sdpPrefix.endsWith("_"))
+	sdpPrefix += "_";
+    for (i = 0; i < n; i++) {
+	const NamedString* param = msg.getParam(i);
+	if (!param)
+	    continue;
+	String tmp = param->name();
+	if (!tmp.startSkip(sdpPrefix,false))
+	    continue;
+	int sep = tmp.find("_");
+	String media("audio");
+	if (sep > 0) {
+	    media = tmp.substr(0,sep);
+	    tmp = tmp.substr(sep + 1);
+	}
+	if (tmp.null() || (tmp.find("_") >= 0))
+	    continue;
+	SDPMedia* rtp = static_cast<SDPMedia*>(m_rtpMedia->operator[](media));
+	if (rtp) {
+	    DDebug(m_enabler,DebugInfo,"Updating %s parameter '%s' to '%s'",
+		media.c_str(),tmp.c_str(),param->c_str());
+	    rtp->parameter(tmp,*param,false);
+	}
+    }
 }
 
 // Add raw SDP forwarding parameter
@@ -945,6 +1011,73 @@ void SDPSession::printRtpMedia(const char* reason)
     Debug(m_enabler,DebugAll,"%s: %s [%p]",reason,tmp.c_str(),m_ptr);
 }
 
+// Set extra parameters for formats
+void SDPSession::setFormatsExtra(const NamedList& list, bool out)
+{
+    const String* amr = list.getParam(out ? YSTRING("oamr_extra") : YSTRING("iamr_extra"));
+    m_amrExtra.assign("");
+    m_amrExtra.clearParams();
+    if (!TelEngine::null(amr)) {
+	ObjList* l = amr->split(';',false);
+	ObjList* o = l->skipNull();
+	if (o) {
+	    m_amrExtra.assign(*amr);
+	    for (; o; o = o->skipNext()) {
+		const String* s = static_cast<String*>(o->get());
+		int pos = s->find('=');
+		if (pos >= 0) {
+		    String n = s->substr(0,pos + 1).trimBlanks();
+		    if (n) {
+			m_amrExtra.addParam(n,s->substr(pos + 1));
+			continue;
+		    }
+		}
+		m_amrExtra.addParam(*s,0);
+	    }
+	}
+	TelEngine::destruct(l);
+    }
+}
+
+// Add extra AMR params to fmtp line
+void SDPSession::addFmtpAmrExtra(String& buf, const String* fmtp)
+{
+    if (TelEngine::null(fmtp)) {
+	buf.append(m_amrExtra,";");
+	return;
+    }
+    // Remove from 'fmtp' parameters we are setting from extra
+    ObjList* l = fmtp->split(';',false);
+    for (ObjList* f = l->skipNull(); f;) {
+	String& s = *static_cast<String*>(f->get());
+	s = s.trimBlanks();
+	NamedString* found = 0;
+	if (s) {
+	    int pos = s.find('=');
+	    if (pos >= 0) {
+		for (ObjList* o = m_amrExtra.paramList()->skipNull(); o; o = o->skipNext()) {
+		    NamedString* ns = static_cast<NamedString*>(o->get());
+		    if (s.startsWith(ns->name())) {
+			found = ns;
+			break;
+		    }
+		}
+	    }
+	    else
+		found = m_amrExtra.getParam(s);
+	}
+	if (found) {
+	    f->remove();
+	    f = f->skipNull();
+	}
+	else
+	    f = f->skipNext();
+    }
+    buf.append(m_amrExtra,";");
+    buf.append(l,";");
+    TelEngine::destruct(l);
+}
+
 };   // namespace TelEngine
 
 /* vi: set ts=8 sw=4 sts=4 noet: */
