@@ -27,10 +27,13 @@
 using namespace TelEngine;
 
 // Constructor from new message
-SIPTransaction::SIPTransaction(SIPMessage* message, SIPEngine* engine, bool outgoing)
+SIPTransaction::SIPTransaction(SIPMessage* message, SIPEngine* engine, bool outgoing,
+    bool* autoChangeParty)
     : m_outgoing(outgoing), m_invite(false), m_transmit(false), m_state(Invalid),
       m_response(0), m_timeouts(0), m_timeout(0),
-      m_firstMessage(message), m_lastMessage(0), m_pending(0), m_engine(engine), m_private(0)
+      m_firstMessage(message), m_lastMessage(0), m_pending(0), m_engine(engine), m_private(0),
+      m_autoChangeParty(autoChangeParty ? *autoChangeParty : engine->autoChangeParty()),
+      m_autoAck(true)
 {
     DDebug(getEngine(),DebugAll,"SIPTransaction::SIPTransaction(%p,%p,%d) [%p]",
 	message,engine,outgoing,this);
@@ -77,7 +80,8 @@ SIPTransaction::SIPTransaction(SIPTransaction& original, SIPMessage* answer)
       m_firstMessage(original.m_firstMessage), m_lastMessage(original.m_lastMessage),
       m_pending(0), m_engine(original.m_engine),
       m_branch(original.m_branch), m_callid(original.m_callid), m_tag(original.m_tag),
-      m_private(0)
+      m_private(0), m_autoChangeParty(original.m_autoChangeParty),
+      m_autoAck(original.m_autoAck)
 {
     DDebug(getEngine(),DebugAll,"SIPTransaction::SIPTransaction(&%p,%p) [%p]",
 	&original,answer,this);
@@ -122,7 +126,8 @@ SIPTransaction::SIPTransaction(const SIPTransaction& original, const String& tag
       m_firstMessage(original.m_firstMessage), m_lastMessage(0),
       m_pending(0), m_engine(original.m_engine),
       m_branch(original.m_branch), m_callid(original.m_callid), m_tag(tag),
-      m_private(0)
+      m_private(0), m_autoChangeParty(original.m_autoChangeParty),
+      m_autoAck(original.m_autoAck)
 {
     if (m_firstMessage)
 	m_firstMessage->ref();
@@ -181,7 +186,7 @@ bool SIPTransaction::changeState(int newstate)
     if ((newstate < 0) || (newstate == m_state))
 	return false;
     if (m_state == Invalid) {
-	Debug(getEngine(),DebugGoOn,"SIPTransaction is already invalid [%p]",this);
+	Debug(getEngine(),DebugWarn,"SIPTransaction is already invalid [%p]",this);
 	return false;
     }
     DDebug(getEngine(),DebugAll,"SIPTransaction state changed from %s to %s [%p]",
@@ -214,7 +219,8 @@ void SIPTransaction::setLatestMessage(SIPMessage* message)
 	m_lastMessage->ref();
 	if (message->isAnswer()) {
 	    m_response = message->code;
-	    if ((m_response > 100) && (m_response < 300))
+	    if ((m_response > 100) && ((m_response < 300)
+		    || !(m_engine->flags() & SIPMessage::NoTagFailure)))
 		setDialogTag();
 	}
 	message->complete(m_engine,0,0,m_tag);
@@ -381,6 +387,29 @@ bool SIPTransaction::setResponse(int code, const char* reason)
     return true;
 }
 
+bool SIPTransaction::setAcknowledge(MimeBody* ackBody)
+{
+    if (!(m_outgoing && m_lastMessage && m_lastMessage->isAnswer()
+	    && ((Process == m_state) || (Retrans == m_state)))) {
+	TelEngine::destruct(ackBody);
+	return false;
+    }
+    m_autoAck = true;
+    // build and send the ACK
+    SIPMessage* m = new SIPMessage(m_firstMessage,m_lastMessage);
+    if (m_autoChangeParty && m_lastMessage->getParty())
+	m->setParty(m_lastMessage->getParty());
+    m->setBody(ackBody);
+    setLatestMessage(m);
+    m_lastMessage->deref();
+    setTransmit();
+    if (changeState(Finish)) {
+	setTimeout(m_engine->getTimer('H'));
+	return true;
+    }
+    return false;
+}
+
 void SIPTransaction::requestAuth(const String& realm, const String& domain, bool stale, bool proxy)
 {
     if (m_outgoing) {
@@ -537,8 +566,15 @@ SIPTransaction::Processed SIPTransaction::processMessage(SIPMessage* message, co
 	}
 	else if (m_tag != *ns) {
 	    // we have a dialog established and this message is out of it
-	    // discriminate forked answers to INVITEs for later processing
-	    return isInvite() ? NoDialog : NoMatch;
+	    if (!isInvite())
+		return NoMatch;
+	    if (message->code < 400)
+		// for 1/2/3xx answers to INVITE discriminate forked for later processing
+		return NoDialog;
+	    // we must answer any 4/5/6xx - and we're supposed to receive only one
+	    DDebug(getEngine(),DebugInfo,"SIPTransaction changing dialog tag '%s' -> '%s' in %u answer [%p]",
+		m_tag.c_str(),ns->c_str(),message->code,this);
+	    m_tag = *ns;
 	}
     }
 
@@ -579,15 +615,12 @@ void SIPTransaction::processClientMessage(SIPMessage* message, int state)
 	    if (final) {
 		setTimeout();
 		if (isInvite()) {
-		    // build the ACK
-		    SIPMessage* m = new SIPMessage(m_firstMessage,message);
-		    if (m_engine->autoChangeParty() && message->getParty())
-			m->setParty(message->getParty());
-		    setLatestMessage(m);
-		    m_lastMessage->deref();
-		    setTransmit();
-		    if (changeState(Finish))
-			setTimeout(m_engine->getTimer('H'));
+		    if ((message->code / 100) != 2)
+			m_autoAck = true;
+		    if (m_autoAck)
+			setAcknowledge();
+		    else
+			changeState(Retrans);
 		}
 		else
 		    changeState(Cleared);
