@@ -5,7 +5,7 @@
  * External module handler
  *
  * Yet Another Telephony Engine - a fully featured software PBX and IVR
- * Copyright (C) 2004-2014 Null Team
+ * Copyright (C) 2004-2017 Null Team
  * Portions copyright (C) 2005 Maciek Kaminski
  *
  * This software is distributed under multiple licenses;
@@ -43,8 +43,12 @@
 using namespace TelEngine;
 namespace { // anonymous
 
-// Maximum length of an incoming line
-#define MAX_INCOMING_LINE 8192
+// Minimum length of the incoming line buffer
+#define MIN_INCOMING_LINE 2048
+// Default length of the incoming line buffer
+#define DEF_INCOMING_LINE 8192
+// Maximum length of the incoming line buffer
+#define MAX_INCOMING_LINE 65536
 
 // Default message timeout in milliseconds
 #define MSG_TIMEOUT 10000
@@ -56,7 +60,7 @@ static Configuration s_cfg;
 static ObjList s_chans;
 static ObjList s_modules;
 static Mutex s_mutex(true,"ExtModule");
-static Mutex s_watchMutex(false,"ExtWatchers");
+static Mutex s_uses(false,"ExtModUse");
 static int s_waitFlush = WAIT_FLUSH;
 static int s_timeout = MSG_TIMEOUT;
 static bool s_timebomb = false;
@@ -191,14 +195,13 @@ public:
     inline MsgWatcher(ExtModReceiver* receiver)
 	: m_receiver(receiver)
 	{ }
-    virtual void destruct();
     virtual void dispatched(const Message& msg, bool handled);
     bool addWatched(const String& name);
     bool delWatched(const String& name);
+    void clear();
 protected:
     virtual void destroyed();
 private:
-    void clear();
     ExtModReceiver* m_receiver;
     ObjList m_watched;
 };
@@ -214,9 +217,10 @@ public:
     };
     static ExtModReceiver* build(const char *script, const char *args, bool ref = false,
 	File* ain = 0, File* aout = 0, ExtModChan *chan = 0);
-    static ExtModReceiver* build(const char* name, Stream* io, ExtModChan* chan = 0, int role = RoleUnknown);
-    static ExtModReceiver* find(const String& script);
-    ~ExtModReceiver();
+    static ExtModReceiver* build(const char* name, Stream* io, ExtModChan* chan = 0,
+	int role = RoleUnknown, const char* conn = 0);
+    static ExtModReceiver* find(const String& script, const String& arg);
+    virtual void destruct();
     virtual bool received(Message& msg, int id);
     bool processLine(const char* line);
     bool outputLine(const char* line);
@@ -229,6 +233,7 @@ public:
     void cleanup();
     bool flush();
     void die(bool clearChan = true);
+    bool useUnlocked();
     bool use();
     bool unuse();
     inline const String& scriptFile() const
@@ -240,13 +245,14 @@ public:
     inline void setRestart(bool restart)
 	{ m_restart = restart; }
     inline bool dead() const
-	{ return m_dead; }
+	{ return m_dead || m_quit || (m_use <= 0); }
     void describe(String& rval) const;
 
 private:
     ExtModReceiver(const char* script, const char* args,
 	File* ain, File* aout, ExtModChan* chan);
-    ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, int role);
+    ExtModReceiver(const char* name, Stream* io, ExtModChan* chan,
+	int role, const char* conn);
     bool create(const char* script, const char* args);
     void closeIn();
     void closeOut();
@@ -254,6 +260,7 @@ private:
     bool outputLineInternal(const char* line, int len);
     int m_role;
     bool m_dead;
+    bool m_quit;
     int m_use;
     pid_t m_pid;
     Stream* m_in;
@@ -269,6 +276,8 @@ private:
     int m_timeout;
     bool m_timebomb;
     bool m_restart;
+    bool m_scripted;
+    DataBlock m_buffer;
     String m_script, m_args;
     ObjList m_waiting;
     ObjList m_relays;
@@ -633,12 +642,6 @@ void ExtMessage::dispatched(bool accepted)
 }
 
 
-void MsgWatcher::destruct()
-{
-    clear();
-    MessagePostHook::destruct();
-}
-
 void MsgWatcher::destroyed()
 {
     clear();
@@ -647,12 +650,15 @@ void MsgWatcher::destroyed()
 
 void MsgWatcher::dispatched(const Message& msg, bool handled)
 {
-    Lock mylock(s_watchMutex);
+    Lock lock(s_uses);
     ExtModReceiver* recv = m_receiver;
-    Lock lock(recv);
-    if (!(recv && recv->use()))
+    if (!recv || recv->dead() || (recv->m_watcher != this) || !recv->useUnlocked())
 	return;
-    mylock.drop();
+    if (!lock.acquire(recv) || !m_receiver || recv->dead()) {
+	lock.drop();
+	recv->unuse();
+	return;
+    }
 
     if (!recv->selfWatch()) {
 	// check if the message was generated by ourselves - avoid reentrance
@@ -701,15 +707,19 @@ bool MsgWatcher::delWatched(const String& name)
 void MsgWatcher::clear()
 {
     Engine::self()->setHook(this,true);
+    if (!m_receiver)
+	return;
+    s_uses.lock();
     ExtModReceiver* recv = m_receiver;
     m_receiver = 0;
     if (recv && (recv->m_watcher == this))
 	recv->m_watcher = 0;
+    s_uses.unlock();
 }
 
 
 ExtModReceiver* ExtModReceiver::build(const char* script, const char* args, bool ref,
-				      File* ain, File* aout, ExtModChan* chan)
+    File* ain, File* aout, ExtModChan* chan)
 {
     ExtModReceiver* recv = new ExtModReceiver(script,args,ain,aout,chan);
     if (ref) {
@@ -723,39 +733,52 @@ ExtModReceiver* ExtModReceiver::build(const char* script, const char* args, bool
     return recv->start() ? recv : 0;
 }
 
-ExtModReceiver* ExtModReceiver::build(const char* name, Stream* io, ExtModChan* chan, int role)
+ExtModReceiver* ExtModReceiver::build(const char* name, Stream* io, ExtModChan* chan,
+    int role, const char* conn)
 {
-    ExtModReceiver* recv = new ExtModReceiver(name,io,chan,role);
+    ExtModReceiver* recv = new ExtModReceiver(name,io,chan,role,conn);
     return recv->start() ? recv : 0;
 }
 
-ExtModReceiver* ExtModReceiver::find(const String& script)
+ExtModReceiver* ExtModReceiver::find(const String& script, const String& arg)
 {
     Lock lock(s_mutex);
     ObjList *l = &s_modules;
     for (; l; l=l->next()) {
 	ExtModReceiver *r = static_cast<ExtModReceiver *>(l->get());
-	if (r && (r->scriptFile() == script))
-	    return r;
+	if (r && (r->scriptFile() == script)) {
+	    if (arg.null() || (r->commandArg() == arg))
+		return r;
+	}
     }
     return 0;
 }
 
+bool ExtModReceiver::useUnlocked()
+{
+    if (m_use <= 0)
+	return false;
+    ++m_use;
+    return true;
+}
+
 bool ExtModReceiver::use()
 {
-    lock();
+    s_uses.lock();
     bool ok = (m_use > 0);
     if (ok)
 	++m_use;
-    unlock();
+    s_uses.unlock();
     return ok;
 }
 
 bool ExtModReceiver::unuse()
 {
-    lock();
-    int u = --m_use;
-    unlock();
+    s_uses.lock();
+    int u = m_use - 1;
+    if (u >= 0)
+	m_use = u;
+    s_uses.unlock();
     if (!u)
 	destruct();
     return (u <= 0);
@@ -763,12 +786,12 @@ bool ExtModReceiver::unuse()
 
 ExtModReceiver::ExtModReceiver(const char* script, const char* args, File* ain, File* aout, ExtModChan* chan)
     : Mutex(true,"ExtModReceiver"),
-      m_role(RoleUnknown), m_dead(false), m_use(1), m_pid(-1),
+      m_role(RoleUnknown), m_dead(false), m_quit(false), m_use(1), m_pid(-1),
       m_in(0), m_out(0), m_ain(ain), m_aout(aout),
       m_chan(chan), m_watcher(0),
       m_selfWatch(false), m_reenter(false), m_setdata(true), m_writing(false),
-      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false),
-      m_script(script), m_args(args), m_trackName(s_trackName)
+      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false), m_scripted(false),
+      m_buffer(0,DEF_INCOMING_LINE), m_script(script), m_args(args), m_trackName(s_trackName)
 {
     Debug(DebugAll,"ExtModReceiver::ExtModReceiver(\"%s\",\"%s\") [%p]",script,args,this);
     m_script.trimBlanks();
@@ -779,17 +802,18 @@ ExtModReceiver::ExtModReceiver(const char* script, const char* args, File* ain,
     s_mutex.unlock();
 }
 
-ExtModReceiver::ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, int role)
+ExtModReceiver::ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, int role, const char* conn)
     : Mutex(true,"ExtModReceiver"),
-      m_role(role), m_dead(false), m_use(1), m_pid(-1),
+      m_role(role), m_dead(false), m_quit(false), m_use(1), m_pid(-1),
       m_in(io), m_out(io), m_ain(0), m_aout(0),
       m_chan(chan), m_watcher(0),
       m_selfWatch(false), m_reenter(false), m_setdata(true), m_writing(false),
-      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false),
-      m_script(name), m_trackName(s_trackName)
+      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false), m_scripted(false),
+      m_buffer(0,DEF_INCOMING_LINE), m_script(name), m_args(conn), m_trackName(s_trackName)
 {
     Debug(DebugAll,"ExtModReceiver::ExtModReceiver(\"%s\",%p,%p) [%p]",name,io,chan,this);
     m_script.trimBlanks();
+    m_args.trimBlanks();
     if (chan)
 	m_role = RoleChannel;
     s_mutex.lock();
@@ -797,10 +821,10 @@ ExtModReceiver::ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, i
     s_mutex.unlock();
 }
 
-ExtModReceiver::~ExtModReceiver()
+void ExtModReceiver::destruct()
 {
-    Debug(DebugAll,"ExtModReceiver::~ExtModReceiver() pid=%d [%p]",m_pid,this);
-    Lock lock(this);
+    Debug(DebugAll,"ExtModReceiver::destruct() pid=%d [%p]",m_pid,this);
+    lock();
     // One destruction is plenty enough
     m_use = -1;
     s_mutex.lock();
@@ -808,7 +832,7 @@ ExtModReceiver::~ExtModReceiver()
     s_mutex.unlock();
     die();
     if (m_pid > 1)
-	Debug(DebugWarn,"ExtModReceiver::~ExtModReceiver() pid=%d [%p]",m_pid,this);
+	Debug(DebugWarn,"ExtModReceiver::destruct() pid=%d [%p]",m_pid,this);
     closeAudio();
     Stream* tmp = m_in;
     m_in = 0;
@@ -818,6 +842,9 @@ ExtModReceiver::~ExtModReceiver()
     tmp = m_out;
     m_out = 0;
     delete tmp;
+    unlock();
+    Thread::yield();
+    GenObject::destruct();
 }
 
 void ExtModReceiver::closeIn()
@@ -864,11 +891,15 @@ bool ExtModReceiver::start()
 
 bool ExtModReceiver::flush()
 {
-    s_watchMutex.lock();
     lock();
-    bool needWait = (0 != m_watcher);
-    TelEngine::destruct(m_watcher);
-    s_watchMutex.unlock();
+    MsgWatcher* w = m_watcher;
+    m_watcher = 0;
+    bool needWait = !!w;
+    if (w) {
+	w->clear();
+	Thread::yield();
+	TelEngine::destruct(w);
+    }
     // Make sure we release all pending messages and not accept new ones
     needWait = needWait || (m_relays.count() != 0);
     if (s_pluginSafe)
@@ -912,6 +943,7 @@ void ExtModReceiver::die(bool clearChan)
     if (m_dead)
 	return;
     m_dead = true;
+    m_quit = true;
     use();
 
     RefPointer<ExtModChan> chan = m_chan;
@@ -920,7 +952,7 @@ void ExtModReceiver::die(bool clearChan)
 	chan->setRecv(0);
     mylock.drop();
 
-    if (m_role == RoleGlobal)
+    if (m_scripted && (m_role == RoleGlobal))
 	Output("Unloading external module '%s' '%s'",m_script.c_str(),m_args.safe());
     // Give the external script a chance to die gracefully
     closeOut();
@@ -954,7 +986,7 @@ void ExtModReceiver::die(bool clearChan)
 
 bool ExtModReceiver::received(Message &msg, int id)
 {
-    if (m_dead)
+    if (m_dead || m_quit)
 	return false;
     lock();
     // check if we are no longer running
@@ -1082,6 +1114,7 @@ bool ExtModReceiver::create(const char *script, const char *args)
     close(ext2yate[1]);
     close(yate2ext[0]);
     closeAudio();
+    m_scripted = true;
     m_pid = pid;
     return true;
 #endif
@@ -1128,14 +1161,14 @@ void ExtModReceiver::run()
     }
     if (m_in && !m_in->setBlocking(false))
 	Debug("ExtModule",DebugWarn,"Failed to set nonblocking mode, expect trouble [%p]",this);
-    char buffer[MAX_INCOMING_LINE];
     int posinbuf = 0;
     bool invalid = true;
     DDebug(DebugAll,"ExtModReceiver::run() entering loop [%p]",this);
     for (;;) {
 	use();
 	lock();
-	int readsize = m_in ? m_in->readData(buffer+posinbuf,sizeof(buffer)-posinbuf-1) : 0;
+	char* buffer = static_cast<char*>(m_buffer.data());
+	int readsize = m_in ? m_in->readData(buffer+posinbuf,m_buffer.length()-posinbuf) : 0;
 	unlock();
 	if (unuse())
 	    return;
@@ -1158,11 +1191,17 @@ void ExtModReceiver::run()
 		Thread::idle();
 		continue;
 	    }
-	    Debug("ExtModule",DebugWarn,"Read error %d on %p [%p]",errno,m_in,this);
+	    if (!m_quit)
+		Debug("ExtModule",DebugWarn,"Read error %d on %p [%p]",errno,m_in,this);
 	    break;
 	}
 	XDebug(DebugAll,"ExtModReceiver::run() read %d",readsize);
 	int totalsize = readsize + posinbuf;
+	if (totalsize >= (int)m_buffer.length()) {
+	    Debug("ExtModule",DebugWarn,"Overflow reading in buffer of length %u, closing [%p]",
+		m_buffer.length(),this);
+	    return;
+	}
 	buffer[totalsize]=0;
 	for (;;) {
 	    char *eoline = ::strchr(buffer,'\n');
@@ -1173,15 +1212,22 @@ void ExtModReceiver::run()
 	    *eoline = 0;
 	    if ((eoline > buffer) && (eoline[-1] == '\r'))
 		eoline[-1] = 0;
+	    readsize = eoline-buffer+1;
 	    if (buffer[0]) {
 		invalid = invalid && (buffer[0] != '%' || buffer[1] != '%');
 		use();
 		bool goOut = processLine(buffer);
 		if (unuse() || goOut)
 		    return;
+		if (totalsize >= (int)m_buffer.length()) {
+		    Debug("ExtModule",DebugWarn,"Lost data shrinking read buffer to %u, closing [%p]",
+			m_buffer.length(),this);
+		    return;
+		}
 	    }
-	    totalsize -= eoline-buffer+1;
-	    ::memmove(buffer,eoline+1,totalsize+1);
+	    totalsize -= readsize;
+	    buffer = static_cast<char*>(m_buffer.data());
+	    ::memmove(buffer,buffer+readsize,totalsize+1);
 	}
 	posinbuf = totalsize;
     }
@@ -1206,8 +1252,9 @@ bool ExtModReceiver::outputLine(const char* line)
 	    break;
 	}
 	if (tout && tout < Time::now()) {
-	    Alarm("extmodule","performance",DebugWarn,"Timeout %d msec for %d characters [%p]",
-		m_timeout,len,this);
+	    if (!m_quit)
+		Alarm("extmodule","performance",DebugWarn,"Timeout %d msec for %d characters [%p]",
+		    m_timeout,len,this);
 	    unuse();
 	    return false;
 	}
@@ -1289,6 +1336,8 @@ bool ExtModReceiver::processLine(const char* line)
 {
     if (m_dead)
 	return false;
+    if (m_quit)
+	return true;
     DDebug("ExtModReceiver",DebugAll,"processLine '%s'", line);
     String id(line);
     if (m_role == RoleUnknown) {
@@ -1355,8 +1404,8 @@ bool ExtModReceiver::processLine(const char* line)
 	static const Regexp r("^\\([^:]*\\):\\([^:]*\\):\\?\\(.*\\)");
 	if (id.matches(r)) {
 	    // a filter is specified
-	    fname = id.matchString(2);
-	    fvalue = id.matchString(3);
+	    fname = String::msgUnescape(id.matchString(2));
+	    fvalue = String::msgUnescape(id.matchString(3));
 	    id = id.matchString(1);
 	}
 	// sanity checks
@@ -1468,8 +1517,18 @@ bool ExtModReceiver::processLine(const char* line)
 		val = m_timebomb;
 		ok = true;
 	    }
+	    else if (id == "bufsize") {
+		unsigned int len = val.toInteger(m_buffer.length(),0,
+		    MIN_INCOMING_LINE,MAX_INCOMING_LINE);
+		if (len > m_buffer.length())
+		    m_buffer.append(DataBlock(0,len - m_buffer.length()));
+		else if (len < m_buffer.length())
+		    m_buffer.assign(m_buffer.data(),len);
+		val = m_buffer.length();
+		ok = true;
+	    }
 	    else if (id == "restart") {
-		m_restart = (RoleGlobal == m_role) && val.toBoolean(m_restart);
+		m_restart = m_scripted && (RoleGlobal == m_role) && val.toBoolean(m_restart);
 		val = m_restart;
 		ok = true;
 	    }
@@ -1527,6 +1586,7 @@ bool ExtModReceiver::processLine(const char* line)
 	}
     }
     else if (id == "%%>quit") {
+	m_quit = true;
 	outputLine("%%<quit");
 	return true;
     }
@@ -1638,7 +1698,7 @@ bool ExtModHandler::received(Message& msg)
     else if (t == "playrec")
 	typ = ExtModChan::DataBoth;
     else {
-	Debug(DebugGoOn,"Invalid ExtModule method '%s', use 'nochan', 'nodata', 'play', 'record' or 'playrec'",
+	Debug(DebugConf,"Invalid ExtModule method '%s', use 'nochan', 'nodata', 'play', 'record' or 'playrec'",
 	    t.c_str());
 	return false;
     }
@@ -1655,7 +1715,7 @@ bool ExtModHandler::received(Message& msg)
     ExtModChan *em = ExtModChan::build(dest.matchString(2).c_str(),
 				       dest.matchString(3).c_str(),typ);
     if (!em) {
-	Debug(DebugGoOn,"Failed to create ExtMod for '%s'",dest.matchString(2).c_str());
+	Debug(DebugCrit,"Failed to create ExtMod for '%s'",dest.matchString(2).c_str());
 	return false;
     }
     ExtModReceiver* recv = em->receiver();
@@ -1714,7 +1774,10 @@ bool ExtModCommand::received(Message& msg)
 	if (line.null())
 	    return false;
 	blank = line.find(' ');
-	ExtModReceiver *r = ExtModReceiver::find(line.substr(0,blank));
+	String arg;
+	if (blank >= 0)
+	    arg = line.substr(blank+1);
+	ExtModReceiver *r = ExtModReceiver::find(line.substr(0,blank),arg);
 	if (r) {
 	    if (start) {
 		msg.retValue() = "External already running\r\n";
@@ -1778,6 +1841,27 @@ bool ExtModCommand::complete(const String& partLine, const String& partWord, Str
 	for (l = mod.skipNull(); l; l = l->skipNext())
 	    Module::itemComplete(rval,l->get()->toString(),partWord);
     }
+    else if (partLine.startsWith("external ")) {
+	String scr = partLine.substr(9);
+	if (!(scr.startSkip("restart") || scr.startSkip("stop")))
+	    return false;
+	if (scr.null() || (scr.find(' ') >= 0))
+	    return false;
+	ObjList arg;
+	s_mutex.lock();
+	ObjList *l = &s_modules;
+	for (; l; l=l->next()) {
+	    ExtModReceiver *r = static_cast<ExtModReceiver *>(l->get());
+	    if (!r || r->commandArg().null() || (r->scriptFile() != scr))
+		continue;
+	    if (arg.find(r->commandArg()))
+		continue;
+	    arg.append(new String(r->commandArg()));
+	}
+	s_mutex.unlock();
+	for (l = arg.skipNull(); l; l = l->skipNext())
+	    Module::itemComplete(rval,l->get()->toString(),partWord);
+    }
     return false;
 }
 
@@ -1882,7 +1966,7 @@ void ExtListener::run()
 	    case ExtModReceiver::RoleUnknown:
 	    case ExtModReceiver::RoleGlobal:
 	    case ExtModReceiver::RoleChannel:
-		ExtModReceiver::build(m_name,skt,0,m_role);
+		ExtModReceiver::build(m_name,skt,0,m_role,tmp);
 		break;
 	    default:
 		Debug(DebugWarn,"Listener '%s' hit invalid role %d",m_name.c_str(),m_role);
