@@ -6,7 +6,7 @@
  * administrating interface.
  *
  * Yet Another Telephony Engine - a fully featured software PBX and IVR
- * Copyright (C) 2004-2014 Null Team
+ * Copyright (C) 2004-2017 Null Team
  *
  * This software is distributed under multiple licenses;
  * see the COPYING file in the main directory for licensing
@@ -161,22 +161,6 @@ static ObjList s_listeners;
 class Connection;
 class RManagerThread;
 
-class SockRef : public RefObject
-{
-public:
-    inline SockRef(Socket** sock)
-	: m_sock(sock)
-	{ }
-    void* getObject(const String& name) const
-    {
-	if (name == YATOM("Socket*"))
-	    return m_sock;
-	return RefObject::getObject(name);
-    }
-private:
-    Socket** m_sock;
-};
-
 class RManagerListener : public RefObject
 {
     friend class RManagerThread;
@@ -224,10 +208,11 @@ public:
     bool processTelnetChar(unsigned char c);
     bool processChar(unsigned char c);
     bool processLine(const char *line, bool saveLine = true);
+    bool processCommand(const char *line, bool saveLine = true);
     bool autoComplete();
     void putConnInfo(NamedList& msg) const;
     void execCommand(const String& str, bool saveOffset);
-    bool pagedCommand(bool up);
+    bool pagedCommand(bool up, bool skip = false);
     void endSubnegotiation();
     void errorBeep();
     void clearLine();
@@ -245,6 +230,7 @@ public:
 	{ return m_listener->cfg(); }
     void checkTimer(u_int64_t time);
 private:
+    void disconnect();
     NamedList m_aliases;
     Level m_auth;
     bool m_debug;
@@ -252,6 +238,8 @@ private:
     bool m_colorize;
     bool m_machine;
     int m_offset;
+    int m_header;
+    int m_finish;
     int m_threshold;
     Socket* m_socket;
     unsigned char m_subBuf[64];
@@ -261,8 +249,10 @@ private:
     unsigned char m_escmode;
     bool m_echoing;
     bool m_beeping;
+    int m_processing;
     u_int64_t m_timeout;
     String m_buffer;
+    String m_prompt;
     String m_address;
     ObjList m_history;
     RefPointer<RManagerListener> m_listener;
@@ -332,13 +322,13 @@ bool RManagerListener::initSocket()
 
     m_socket.create(AF_INET, SOCK_STREAM);
     if (!m_socket.valid()) {
-	Alarm("RManager","socket",DebugGoOn,"Unable to create the listening socket: %s",
+	Alarm("RManager","socket",DebugCrit,"Unable to create the listening socket: %s",
 	    strerror(m_socket.error()));
 	return false;
     }
 
     if (!m_socket.setBlocking(false)) {
-	Alarm("RManager","socket",DebugGoOn, "Failed to set listener to nonblocking mode: %s",
+	Alarm("RManager","socket",DebugCrit, "Failed to set listener to nonblocking mode: %s",
 	    strerror(m_socket.error()));
 	return false;
     }
@@ -349,12 +339,12 @@ bool RManagerListener::initSocket()
     m_address << sa.host() << ":" << sa.port();
     m_socket.setReuse();
     if (!m_socket.bind(sa)) {
-	Alarm("RManager","socket",DebugGoOn,"Failed to bind to %s : %s",
+	Alarm("RManager","socket",DebugCrit,"Failed to bind to %s : %s",
 	    m_address.c_str(),strerror(m_socket.error()));
 	return false;
     }
     if (!m_socket.listen(2)) {
-	Alarm("RManager","socket",DebugGoOn,"Unable to listen on socket: %s",
+	Alarm("RManager","socket",DebugCrit,"Unable to listen on socket: %s",
 	    strerror(m_socket.error()));
 	return false;
     }
@@ -403,7 +393,7 @@ Connection* RManagerListener::checkCreate(Socket* sock, const char* addr)
 	m.addParam("server",String::boolText(true));
 	m.addParam(secure->name(),*secure);
 	m.copyParam(m_cfg,"verify");
-	SockRef* s = new SockRef(&sock);
+	SocketRef* s = new SocketRef(sock);
 	m.userData(s);
 	TelEngine::destruct(s);
 	if (!(Engine::dispatch(m) && sock)) {
@@ -414,14 +404,15 @@ Connection* RManagerListener::checkCreate(Socket* sock, const char* addr)
 	}
     }
     else if (!sock->setBlocking(false)) {
-	Debug("RManager",DebugGoOn, "Failed to set tcp socket to nonblocking mode: %s",
+	Debug("RManager",DebugCrit, "Failed to set tcp socket to nonblocking mode: %s",
 	    strerror(sock->error()));
 	delete sock;
 	return 0;
     }
     // should check IP address here
-    Output("Remote%s connection from %s to %s",
-	(secure ? " secure" : ""),addr,m_address.c_str());
+    if (m_cfg.getBoolValue("logged",true))
+	Output("Remote%s connection from %s to %s",
+	    (secure ? " secure" : ""),addr,m_address.c_str());
     Connection* conn = new Connection(sock,addr,this);
     if (conn->error()) {
 	conn->destruct();
@@ -436,12 +427,16 @@ Connection::Connection(Socket* sock, const char* addr, RManagerListener* listene
     : Thread("RManager Connection"),
       m_aliases(""),
       m_auth(None), m_debug(false), m_output(false), m_colorize(false), m_machine(false),
-      m_offset(-1), m_threshold(DebugAll),
+      m_offset(-1), m_header(0), m_finish(-1), m_threshold(DebugAll),
       m_socket(sock), m_subOpt(0), m_subLen(0),
-      m_lastch(0), m_escmode(0), m_echoing(false), m_beeping(false),
+      m_lastch(0), m_escmode(0), m_echoing(false), m_beeping(false), m_processing(0),
       m_timeout(0), m_address(addr), m_listener(listener),
       m_cursorPos(0), m_histLen(DEF_HISTORY), m_width(0), m_height(24)
 {
+    m_prompt = cfg().getValue("prompt");
+    if ((m_prompt.length() > 1) && (m_prompt.startsWith("\"")) && (m_prompt.endsWith("\"")))
+	m_prompt = m_prompt.substr(1,m_prompt.length() - 2);
+    Engine::runParams().replaceParams(m_prompt);
     s_mutex.lock();
     s_connList.append(this);
     s_mutex.unlock();
@@ -454,9 +449,18 @@ Connection::~Connection()
     s_mutex.lock();
     s_connList.remove(this,false);
     s_mutex.unlock();
-    Output("Closing connection to %s",m_address.c_str());
-    delete m_socket;
+    if (cfg().getBoolValue("logged",true))
+	Output("Closing connection to %s",m_address.c_str());
+    Socket* tmp = m_socket;
     m_socket = 0;
+    yield();
+    delete tmp;
+}
+
+void Connection::disconnect()
+{
+    if (m_socket)
+	m_socket->terminate();
 }
 
 void Connection::run()
@@ -480,9 +484,9 @@ void Connection::run()
     }
     else {
 	m_auth = cfg().getValue("password") ? User : Admin;
-	m_output = cfg().getBoolValue("output",false);
 	if (Admin == m_auth)
 	    m_debug = cfg().getBoolValue("debug",false);
+	m_output = cfg().getBoolValue("output",m_debug);
     }
     m_histLen = cfg().getIntValue("maxhistory",DEF_HISTORY);
     if (m_histLen > MAX_HISTORY)
@@ -494,10 +498,14 @@ void Connection::run()
 	// WILL SUPPRESS GO AHEAD, WILL ECHO, DO NAWS - and enough BS and blanks to hide them
 	writeStr("\377\373\003\377\373\001\377\375\037\r         \b\b\b\b\b\b\b\b\b");
     }
+    else
+	m_echoing = cfg().getBoolValue("echoing",false);
     if (hdr) {
 	writeStr("\r" + hdr + "\r\n");
 	hdr.clear();
     }
+    if (m_prompt && m_echoing)
+	writeBuffer();
     NamedIterator iter(cfg());
     while (const NamedString* s = iter.get()) {
 	if (s->null() || !s->name().startsWith("alias:"))
@@ -511,18 +519,21 @@ void Connection::run()
 	Thread::check();
 	bool readok = false;
 	bool error = false;
+	SOCKET sh = m_socket->handle();
 	if (m_socket->select(&readok,0,&error,10000)) {
 	    // rearm the error beep
 	    m_beeping = false;
 	    if (error) {
-		Debug("RManager",DebugInfo,"Socket exception condition on %d",m_socket->handle());
+		disconnect();
+		Debug("RManager",DebugInfo,"Socket exception condition on %d",sh);
 		return;
 	    }
 	    if (!readok)
 		continue;
 	    int readsize = m_socket->readData(buffer,sizeof(buffer));
 	    if (!readsize) {
-		Debug("RManager",DebugInfo,"Socket condition EOF on %d",m_socket->handle());
+		disconnect();
+		Debug("RManager",DebugInfo,"Socket condition EOF on %d",sh);
 		return;
 	    }
 	    else if (readsize > 0) {
@@ -531,12 +542,14 @@ void Connection::run()
 			return;
 	    }
 	    else if (!m_socket->canRetry()) {
-		Debug("RManager",DebugWarn,"Socket read error %d on %d",errno,m_socket->handle());
+		disconnect();
+		Debug("RManager",DebugWarn,"Socket read error %d on %d",errno,sh);
 		return;
 	    }
 	}
 	else if (!m_socket->canRetry()) {
-	    Debug("RManager",DebugWarn,"socket select error %d on %d",errno,m_socket->handle());
+	    disconnect();
+	    Debug("RManager",DebugWarn,"socket select error %d on %d",errno,sh);
 	    return;
 	}
     }
@@ -572,12 +585,14 @@ void Connection::writeBufferTail(bool eraseOne)
 // write the current buffer, leave the cursor in the right place
 void Connection::writeBuffer()
 {
+    if (m_processing)
+	return;
     if (m_cursorPos == m_buffer.length()) {
-	writeStr(m_buffer);
+	writeStr(m_prompt + m_buffer);
 	return;
     }
-    if (m_cursorPos)
-	writeStr(m_buffer,m_cursorPos);
+    if (m_cursorPos || m_prompt)
+	writeStr(m_prompt + m_buffer,m_prompt.length() + m_cursorPos);
     writeBufferTail();
 }
 
@@ -652,7 +667,7 @@ bool Connection::processTelnetChar(unsigned char c)
 		switch (c) {
 		    case 1: // ECHO
 			m_echoing = true;
-			writeStr("\377\373\001"); // WILL ECHO
+			writeStr("\377\373\001" + m_prompt); // WILL ECHO
 			break;
 		    case 3: // SUPPRESS GO AHEAD
 			writeStr("\377\373\003"); // WILL SUPPRESS GO AHEAD
@@ -729,10 +744,14 @@ bool Connection::processChar(unsigned char c)
 	    m_escmode = 0;
 	    if (m_echoing)
 		writeStr("\r\n");
-	    if (processLine(m_buffer))
-		return true;
-	    m_buffer.clear();
-	    m_cursorPos = 0;
+	    if (m_buffer) {
+		if (processLine(m_buffer))
+		    return true;
+		m_cursorPos = 0;
+		m_buffer.clear();
+	    }
+	    if (m_echoing && m_prompt)
+		writeBuffer();
 	    return false;
 	case 0x03: // ^C, BREAK
 	    m_escmode = 0;
@@ -744,10 +763,14 @@ bool Connection::processChar(unsigned char c)
 		errorBeep();
 		return false;
 	    }
+	    if (m_echoing && m_prompt)
+		writeStr("\r\n");
 	    return processLine("quit",false);
 	case 0x1C: // ^backslash
 	    if (m_buffer)
 		break;
+	    if (m_echoing && m_prompt)
+		writeStr("\r\n");
 	    return processLine("reload",false);
 	case 0x05: // ^E
 	    m_escmode = 0;
@@ -759,13 +782,14 @@ bool Connection::processChar(unsigned char c)
 		errorBeep();
 		return false;
 	    }
-	    // cycle [no output] -> [output] -> [debug (only if admin)]
+	    // cycle [no output] -> [output] -> [output+debug (only if admin)]
 	    if (m_debug)
 		m_output = m_debug = false;
 	    else if (m_output) {
-		m_output = false;
 		if ((m_debug = (m_auth >= Admin)))
 		    Debugger::enableOutput(true);
+		else
+		    m_output = false;
 	    }
 	    else
 		m_output = true;
@@ -786,10 +810,12 @@ bool Connection::processChar(unsigned char c)
 	    if (m_buffer.null())
 		break;
 	    m_escmode = 0;
-	    m_buffer.clear();
 	    m_cursorPos = 0;
-	    if (m_echoing)
+	    m_buffer.clear();
+	    if (m_echoing) {
 		clearLine();
+		writeBuffer();
+	    }
 	    return false;
 	case 0x17: // ^W
 	    if (m_cursorPos <= 0)
@@ -836,8 +862,12 @@ bool Connection::processChar(unsigned char c)
 	    return false;
 	case 0x09: // ^I, TAB
 	    m_escmode = 0;
-	    if (m_buffer.null())
-		return processLine("help",false);
+	    if (m_buffer.null()) {
+		processLine("help",false);
+		if (m_echoing)
+		    writeBuffer();
+		return false;
+	    }
 	    if (!atEol) {
 		if (m_echoing)
 		    writeStr(m_buffer.c_str()+m_cursorPos,m_buffer.length()-m_cursorPos);
@@ -881,8 +911,8 @@ bool Connection::processChar(unsigned char c)
 		    m_buffer = s;
 		    TelEngine::destruct(s);
 		}
-		clearLine();
 		m_cursorPos = m_buffer.length();
+		clearLine();
 		writeBuffer();
 		return false;
 	    case 'B': // Down arrow
@@ -897,8 +927,8 @@ bool Connection::processChar(unsigned char c)
 		    m_buffer = s;
 		    TelEngine::destruct(s);
 		}
-		clearLine();
 		m_cursorPos = m_buffer.length();
+		clearLine();
 		writeBuffer();
 		return false;
 	    case 'C': // Right arrow
@@ -920,29 +950,37 @@ bool Connection::processChar(unsigned char c)
 		m_cursorPos--;
 		return false;
 	    case 'H': // Home
-		if (m_echoing)
-		    writeStr("\r");
+		if (pagedCommand(true,true))
+		    return false;
 		m_cursorPos = 0;
+		if (m_echoing)
+		    writeStr("\r" + m_prompt);
 		return false;
 	    case 'F': // End
+		if (pagedCommand(false,true))
+		    return false;
 		if (atEol)
 		    return false;
 		if (m_echoing && m_buffer)
-		    writeStr(m_buffer.c_str()+m_cursorPos);
+		    writeStr(m_buffer.c_str() + m_cursorPos);
 		m_cursorPos = m_buffer.length();
 		return false;
 	    case '~':
 		switch (escMode) {
 		    case '1': // Home
-			if (m_echoing)
-			    writeStr("\r");
+			if (pagedCommand(true,true))
+			    return false;
 			m_cursorPos = 0;
+			if (m_echoing)
+			    writeStr("\r" + m_prompt);
 			return false;
 		    case '4': // End
+			if (pagedCommand(false,true))
+			    return false;
 			if (atEol)
 			    return false;
 			if (m_echoing && m_buffer)
-			    writeStr(m_buffer.c_str()+m_cursorPos);
+			    writeStr(m_buffer.c_str() + m_cursorPos);
 			m_cursorPos = m_buffer.length();
 			return false;
 		    case '3': // Delete
@@ -997,25 +1035,44 @@ bool Connection::processChar(unsigned char c)
     return false;
 }
 
-bool Connection::pagedCommand(bool up)
+bool Connection::pagedCommand(bool up, bool skip)
 {
-    if (m_buffer || m_offset < 0)
+    if (m_buffer || m_offset <= 0)
 	return false;
     const String* s = static_cast<const String*>(m_history.get());
     if (!s)
 	return false;
     if (up) {
-	if (!m_height)
-	    return false;
-	int offs = m_offset - m_height;
-	if (offs < 0)
-	    return false;
-	offs -= m_height;
-	if (offs < 0)
-	    offs = 0;
-	m_offset = offs;
+	if (skip)
+	    m_offset = 0;
+	else {
+	    if (!m_height)
+		return false;
+	    int offs = m_offset - m_height + m_header;
+	    if (!offs)
+		return false;
+	    offs -= m_height - m_header;
+	    if (offs < 0)
+		offs = 0;
+	    m_offset = offs;
+	}
     }
+    else if (m_finish >= 0) {
+	if (m_offset == m_finish)
+	    return false;
+	int last = m_finish - m_height + m_header;
+	if (last < 0)
+	    return false;
+	if (skip || (m_offset > last))
+	    m_offset = last;
+    }
+    if (m_echoing && m_prompt)
+	writeStr("\r\n");
+    m_processing++;
     execCommand(*s,true);
+    m_processing--;
+    if (m_echoing && m_prompt)
+	writeBuffer();
     return true;
 }
 
@@ -1027,7 +1084,7 @@ void Connection::putConnInfo(NamedList& msg) const
     msg.setParam("cmd_machine",String::boolText(m_machine));
     if (m_width && m_height && !m_machine) {
 	msg.setParam("cmd_width",String(m_width));
-	msg.setParam("cmd_heigth",String(m_height));
+	msg.setParam("cmd_height",String(m_height));
     }
 }
 
@@ -1189,10 +1246,19 @@ bool Connection::autoComplete()
     return true;
 }
 
-// execute received input line
+// process received input line
 bool Connection::processLine(const char *line, bool saveLine)
 {
-    DDebug("RManager",DebugInfo,"processLine = '%s'",line);
+    m_processing++;
+    bool ok = processCommand(line,saveLine);
+    m_processing--;
+    return ok;
+}
+
+// execute received command
+bool Connection::processCommand(const char *line, bool saveLine)
+{
+    DDebug("RManager",DebugInfo,"processCommand = '%s'",line);
     String str(line);
     str.trimBlanks();
     if (str.null())
@@ -1208,10 +1274,21 @@ bool Connection::processLine(const char *line, bool saveLine)
     line = 0;
     m_buffer.clear();
     m_offset = -1;
+    m_header = 0;
+    m_finish = -1;
 
     if (str.startSkip("quit"))
     {
-	writeStr(m_machine ? "%%=quit\r\n" : "Goodbye!\r\n");
+	if (m_machine)
+	    writeStr("%%=quit\r\n");
+	else {
+	    String txt(cfg().getValue("goodbye","Goodbye!"));
+	    Engine::runParams().replaceParams(txt);
+	    if (txt) {
+		txt << "\r\n";
+		writeStr(txt);
+	    }
+	}
 	return true;
     }
     else if (str.startSkip("echo"))
@@ -1314,9 +1391,13 @@ bool Connection::processLine(const char *line, bool saveLine)
 	}
 	putConnInfo(m);
 	Engine::dispatch(m);
-	str = "%%+status" + str + "\r\n";
-	str << m.retValue() << "%%-status\r\n";
-	writeStr(str);
+	if (m_echoing && m_prompt)
+	    writeStr(m.retValue());
+	else {
+	    str = "%%+status" + str + "\r\n";
+	    str << m.retValue() << "%%-status\r\n";
+	    writeStr(str);
+	}
 	return false;
     }
     else if (str.startSkip("uptime"))
@@ -1428,6 +1509,14 @@ bool Connection::processLine(const char *line, bool saveLine)
 	if (str.startSkip("level")) {
 	    int dbg = debugLevel();
 	    str >> dbg;
+	    if (str.startSkip("+")) {
+		if (debugLevel() > dbg)
+		    dbg = debugLevel();
+	    }
+	    else if (str.startSkip("-")) {
+		if (debugLevel() < dbg)
+		    dbg = debugLevel();
+	    }
 	    debugLevel(dbg);
 	}
 	NamedCounter* counter = 0;
@@ -1678,8 +1767,8 @@ bool Connection::processLine(const char *line, bool saveLine)
 	    for (;;) {
 		sep = str.find("$()");
 		if (sep < 0)
-		    return processLine(str,false);
-		if (processLine(str.substr(0,sep),false))
+		    return processCommand(str,false);
+		if (processCommand(str.substr(0,sep),false))
 		    return true;
 		str = str.substr(sep+3);
 	    }
@@ -1697,8 +1786,12 @@ void Connection::execCommand(const String& str, bool saveOffset)
     Message m("engine.command");
     m.addParam("line",str);
     putConnInfo(m);
-    if (saveOffset && m_offset >= 0)
+    if (saveOffset && m_offset >= 0) {
 	m.setParam("cmd_offset",String(m_offset));
+	m.setParam("cmd_header",String(m_header));
+	if (m_finish >= 0)
+	    m.setParam("cmd_finish",String(m_finish));
+    }
     if (Engine::dispatch(m)) {
 	writeStr(m.retValue());
 	const ObjList* l = YOBJECT(ObjList,m.userData());
@@ -1712,8 +1805,11 @@ void Connection::execCommand(const String& str, bool saveOffset)
 	    else
 		writeEvent(o->toString(),-1);
 	}
-	if (saveOffset)
+	if (saveOffset) {
 	    m_offset = m.getIntValue("cmd_offset",-1);
+	    m_header = m.getIntValue("cmd_header",0);
+	    m_finish = m.getIntValue("cmd_finish",m_finish);
+	}
     }
     else
 	writeStr((m_machine ? "%%=syntax:" : "Cannot understand: ") + str + "\r\n");
@@ -1726,17 +1822,18 @@ void Connection::writeStr(const Message &msg, bool received)
 	return;
     String s = msg.encode(received,"");
     s << "\r\n";
-    if (m_echoing && m_buffer)
+    bool dirty = m_echoing && (m_buffer || m_prompt) && !m_processing;
+    if (dirty)
 	clearLine();
     writeStr(s.c_str());
-    if (m_echoing && m_buffer)
+    if (dirty)
 	writeBuffer();
 }
 
 // write debugging messages to the remote console
 void Connection::writeDebug(const char *str, int level)
 {
-    if ((m_debug && (m_threshold >= level)) || (m_output && (level < 0)))
+    if ((level < 0) ? m_output : (m_debug && (m_threshold >= level)))
 	writeEvent(str,level);
 }
 
@@ -1745,7 +1842,8 @@ void Connection::writeEvent(const char *str, int level)
 {
     if (null(str))
 	return;
-    if (m_echoing && m_buffer)
+    bool dirty = m_echoing && (m_buffer || m_prompt) && !m_processing;
+    if (dirty)
 	clearLine();
     const char* col = m_colorize ? debugColor(level) : 0;
     if (col)
@@ -1761,7 +1859,7 @@ void Connection::writeEvent(const char *str, int level)
 	col = debugColor(-2);
     if (col)
 	writeStr(col,::strlen(col));
-    if (m_echoing && m_buffer)
+    if (dirty)
 	writeBuffer();
 }
 
@@ -1772,8 +1870,12 @@ void Connection::writeStr(const char *str, int len)
 	len = ::strlen(str);
     else if (len == 0)
 	return;
+    if (!(m_socket && m_socket->valid()))
+	return;
+    SOCKET sh = m_socket->handle();
     if (int written = m_socket->writeData(str,len) != len) {
-	Debug("RManager",DebugInfo,"Socket %d wrote only %d out of %d bytes",m_socket->handle(),written,len);
+	disconnect();
+	Debug("RManager",DebugInfo,"Socket %d wrote only %d out of %d bytes",sh,written,len);
 	// Destroy the thread, will kill the connection
 	cancel();
     }
@@ -1786,8 +1888,7 @@ void Connection::checkTimer(u_int64_t time)
     if (time < m_timeout)
 	return;
     m_timeout = 0;
-    if (m_socket)
-	m_socket->terminate();
+    disconnect();
 }
 
 
