@@ -281,9 +281,10 @@ void MyConn::closeConn()
     MYSQL* tmp = m_conn;
     m_conn = 0;
     mysql_close(tmp);
+    String name(*this);
     if (m_owner)
 	m_owner->m_connections.remove(this);
-    Debug(&module,DebugInfo,"Database connection '%s' closed",c_str());
+    Debug(&module,DebugInfo,"Database connection '%s' closed",name.c_str());
 }
 
 void MyConn::runQueries()
@@ -365,7 +366,7 @@ int MyConn::queryDbInternal(DbQuery* query)
 		    if (columns[c])
 			columns[c]->set(new String(fields[c].name));
 		    else
-			Debug(&module,DebugGoOn,"No array for column %u",c);
+			Debug(&module,DebugCrit,"No array for column %u",c);
 		}
 		// and now data row by row
 		for (unsigned int r = 1; r <= rows; r++) {
@@ -479,7 +480,7 @@ int MyAcct::initConns()
 
 	mySqlConn->m_conn = mysql_init(mySqlConn->m_conn);
 	if (!mySqlConn->m_conn) {
-	    Debug(&module,DebugGoOn,"Could not start connection %d for '%s'",i,c_str());
+	    Debug(&module,DebugCrit,"Could not start connection %d for '%s'",i,c_str());
 	    return i;
 	}
 	DDebug(&module,DebugAll,"Connection '%s' for account '%s' was created",mySqlConn->c_str(),c_str());
